"""
The following code is a modification of a hierarchical losses implementation first published at https://github.com/jvlmdr/hiercls
"""

from functools import partial
import numpy as np
import collections
import csv
import itertools
from typing import Callable, Hashable, Optional, Sequence, Tuple, TextIO, cast

import networkx as nx
from numpy.typing import ArrayLike

import torch
from torch import nn
import torch.nn.functional as F


class Hierarchy:
    """Hierarchy of nodes 0, ..., n-1."""

    def __init__(self, parents):
        n = len(parents)
        assert np.all(parents[1:] < np.arange(1, n))
        self._parents = parents

    def num_nodes(self) -> int:
        return len(self._parents)

    def edges(self) -> list[tuple[int, int]]:
        return list(zip(self._parents[1:], itertools.count(1)))

    def parents(self, root_loop: bool = False) -> np.ndarray:
        if root_loop:
            return np.where(
                np.array(self._parents) >= 0,
                np.array(self._parents),
                np.arange(len(self._parents)),
            )
        else:
            return np.array(self._parents)

    def children(self) -> dict[int, np.ndarray]:
        result = collections.defaultdict(list)
        for i, j in self.edges():
            result[i].append(j)
        return {k: np.array(v, dtype=int) for k, v in result.items()}

    def num_children(self) -> np.ndarray:
        n = len(self._parents)
        unique, counts = np.unique(self._parents[1:], return_counts=True)
        result = np.zeros([n], dtype=int)
        result[unique] = counts
        return result

    def leaf_mask(self) -> np.ndarray:
        return self.num_children() == 0

    def leaf_subset(self) -> np.ndarray:
        (index,) = self.leaf_mask().nonzero()
        return index

    def internal_subset(self) -> np.ndarray:
        (index,) = np.logical_not(self.leaf_mask()).nonzero()
        return index

    def num_leaf_nodes(self) -> int:
        return np.count_nonzero(self.leaf_mask())

    def num_internal_nodes(self) -> int:
        return np.count_nonzero(np.logical_not(self.leaf_mask()))

    def num_conditionals(self) -> int:
        return np.count_nonzero(self.num_children() > 1)

    def depths(self) -> np.ndarray:
        return self.accumulate_ancestors(np.add, (self._parents >= 0).astype(int))

    def num_leaf_descendants(self) -> np.ndarray:
        return self.accumulate_descendants(np.add, self.leaf_mask().astype(int))

    def max_heights(self) -> np.ndarray:
        heights = np.zeros_like(self.depths())
        return self.accumulate_descendants(lambda u, v: max(u, v + 1), heights)

    def min_heights(self) -> np.ndarray:
        # Initialize leaf nodes to zero, internal nodes to upper bound.
        depths = self.depths()
        heights = np.where(self.leaf_mask(), 0, depths.max() - depths)
        return self.accumulate_descendants(lambda u, v: min(u, v + 1), heights)

    def accumulate_ancestors(self, func: Callable, values: ArrayLike) -> np.ndarray:
        # Start from root and move down.
        partials = np.array(values)
        for i, j in self.edges():
            partials[j] = func(partials[i], partials[j])
        return partials

    def accumulate_descendants(self, func: Callable, values: ArrayLike) -> np.ndarray:
        # Start from leaves and move up.
        partials = np.array(values)
        for i, j in reversed(self.edges()):
            partials[i] = func(partials[i], partials[j])
        return partials

    def ancestor_mask(self, strict=False) -> np.ndarray:
        n = len(self._parents)
        # If path exists i, ..., j then i < j and is_ancestor[i, j] is True.
        # Work with is_descendant instead to use consecutive blocks of memory.
        # Note that is_ancestor[i, j] == is_descendant[j, i].
        is_descendant = np.zeros([n, n], dtype=bool)
        if not strict:
            is_descendant[0, 0] = 1
        for i, j in self.edges():
            # Node i is parent of node j.
            assert i < j, "require edges in topological order"
            is_descendant[j, :] = is_descendant[i, :]
            if strict:
                is_descendant[j, i] = 1
            else:
                is_descendant[j, j] = 1
        is_ancestor = is_descendant.T
        return is_ancestor

    def paths(
        self,
        exclude_root: bool = False,
        exclude_self: bool = False,
    ) -> list[np.ndarray]:
        is_descendant = self.ancestor_mask(strict=exclude_self).T
        if exclude_root:
            paths = [np.flatnonzero(mask) + 1 for mask in is_descendant[:, 1:]]
        else:
            paths = [np.flatnonzero(mask) for mask in is_descendant]
        return paths

    def paths_padded(
        self, pad_value=-1, method: str = "constant", **kwargs
    ) -> np.ndarray:
        n = self.num_nodes()
        paths = self.paths(**kwargs)
        path_lens = list(map(len, paths))
        max_len = max(path_lens)
        if method == "constant":
            padded = np.full((n, max_len), pad_value, dtype=int)
        elif method == "self":
            padded = np.tile(np.arange(n)[:, None], max_len)
        else:
            raise ValueError("unknown pad method", method)
        row_index = np.concatenate([np.full(n, i) for i, n in enumerate(path_lens)])
        col_index = np.concatenate([np.arange(n) for n in path_lens])
        padded[row_index, col_index] = np.concatenate(paths)
        return padded

    def __str__(self, node_names: Optional[list[str]] = None) -> str:
        return format_tree(self, node_names)

    def to_networkx(self, keys: Optional[list] = None) -> nx.DiGraph:
        n = self.num_nodes()
        g = nx.DiGraph()
        if keys is None:
            keys = list(range(n))
        g.add_edges_from([(keys[i], keys[j]) for i, j in self.edges()])
        return g


def make_hierarchy_from_edges(
    pairs: Sequence[tuple[str, str]],
) -> tuple[Hierarchy, list[str]]:
    """Creates a hierarchy from a list of name pairs.

    The order of the edges determines the order of the nodes.
    (Each node except the root appears once as a child.)
    The root is placed first in the order.
    """
    num_edges = len(pairs)
    num_nodes = num_edges + 1
    # Data structures to populate from list of pairs.
    parents = np.full([num_nodes], -1, dtype=int)
    names: list[Optional[str]] = [None] * num_nodes
    name_to_index = {}
    # Set name of root from first pair.
    root, _ = pairs[0]
    names[0] = root
    name_to_index[root] = 0
    for r, (u, v) in enumerate(pairs):
        if v in name_to_index:
            raise ValueError("has multiple parents", v)
        i = name_to_index[u]
        j = r + 1
        parents[j] = i
        names[j] = v
        name_to_index[v] = j
    assert all(i is not None for i in names)
    return Hierarchy(parents), cast(list[str], names)


def load_edges(f: TextIO, delimiter=",") -> list[tuple[str, str]]:
    """Load from file containing (parent, node) pairs."""
    reader = csv.reader(f)
    pairs = []
    for row in reader:
        if not row:
            continue
        if len(row) != 2:
            raise ValueError("invalid row", row)
        pairs.append(tuple(row))
    return pairs


def rooted_subtree(tree: Hierarchy, nodes: np.ndarray) -> Hierarchy:
    """Finds the subtree that contains a subset of nodes."""
    # Check that root is present in subset.
    assert nodes[0] == 0
    # Construct a new list of parents.
    reindex = np.full([tree.num_nodes()], -1)
    reindex[nodes] = np.arange(len(nodes))
    parents = tree.parents()
    subtree_parents = np.where(parents[nodes] >= 0, reindex[parents[nodes]], -1)
    assert np.all(subtree_parents[1:] >= 0), "parent not in subset"
    # Ensure that parent appears before child.
    assert np.all(subtree_parents < np.arange(len(nodes)))
    return Hierarchy(subtree_parents)


def rooted_subtree_spanning(
    tree: Hierarchy, nodes: np.ndarray
) -> tuple[Hierarchy, np.ndarray]:
    nodes = ancestors_union(tree, nodes)
    subtree = rooted_subtree(tree, nodes)
    return subtree, nodes


def ancestors_union(tree: Hierarchy, node_subset: np.ndarray) -> np.ndarray:
    """Returns union of ancestors of nodes."""
    paths = tree.paths_padded(-1)
    paths = paths[node_subset]
    return np.unique(paths[paths >= 0])


def find_subset_index(base: list[Hashable], subset: list[Hashable]) -> np.ndarray:
    """Returns index of subset elements in base list (injective map)."""
    name_to_index = {x: i for i, x in enumerate(base)}
    return np.asarray([name_to_index[x] for x in subset], dtype=int)


def find_projection(tree: Hierarchy, node_subset: np.ndarray) -> np.ndarray:
    """Finds projection to nearest ancestor in subtree."""
    # Use paths rather than ancestor_mask to avoid large memory usage.
    assert np.all(node_subset >= 0)
    paths = tree.paths_padded(-1)
    # Find index in subset.
    reindex = np.full([tree.num_nodes()], -1)
    reindex[node_subset] = np.arange(len(node_subset))
    subset_paths = np.where(paths >= 0, reindex[paths], -1)
    deepest = _last_nonzero(subset_paths >= 0, axis=1)
    # Check that all ancestors are present.
    assert np.all(np.count_nonzero(subset_paths >= 0, axis=1) - 1 == deepest)
    return subset_paths[np.arange(tree.num_nodes()), deepest]


def _last_nonzero(x, axis):
    x = np.asarray(x, bool)
    assert np.all(np.any(x, axis=axis)), "at least one must be nonzero"
    # Find last element that is true.
    # (First element that is true in reversed array.)
    n = x.shape[axis]
    return (n - 1) - np.argmax(np.flip(x, axis), axis=axis)


def uniform_leaf(tree: Hierarchy) -> np.ndarray:
    """Returns a uniform distribution over leaf nodes."""
    is_ancestor = tree.ancestor_mask(strict=False)
    is_leaf = tree.leaf_mask()
    num_leaf_descendants = is_ancestor[:, is_leaf].sum(axis=1)
    return num_leaf_descendants / is_leaf.sum()


def uniform_cond(tree: Hierarchy) -> np.ndarray:
    """Returns a uniform distribution over child nodes at each conditional."""
    node_to_num_children = {k: len(v) for k, v in tree.children().items()}
    num_children = np.asarray(
        [node_to_num_children.get(x, 0) for x in range(tree.num_nodes())]
    )
    parent_index = tree.parents()
    # Root node has likelihood 1 and no parent.
    log_cond_p = np.concatenate([[0.0], -np.log(num_children[parent_index[1:]])])
    is_ancestor = tree.ancestor_mask(strict=False)
    log_p = np.dot(is_ancestor.T, log_cond_p)
    return np.exp(log_p)


def lca_depth(tree: Hierarchy, inds_a: np.ndarray, inds_b: np.ndarray) -> np.ndarray:
    """Returns the depth of the LCA node.

    Supports multi-dimensional index arrays.
    """
    paths = tree.paths_padded(exclude_root=True)
    paths_a = paths[inds_a]
    paths_b = paths[inds_b]
    return np.count_nonzero(
        ((paths_a == paths_b) & (paths_a >= 0) & (paths_b >= 0)), axis=-1
    )


def find_lca(tree: Hierarchy, inds_a: np.ndarray, inds_b: np.ndarray) -> np.ndarray:
    """Returns the index of the LCA node.

    Supports multi-dimensional index arrays.
    For example, to obtain an exhaustive table:
        n = tree.num_nodes()
        find_lca(tree, np.arange(n)[:, np.newaxis], np.arange(n)[np.newaxis, :])
    """
    paths = tree.paths_padded(exclude_root=False)
    paths_a = paths[inds_a]
    paths_b = paths[inds_b]
    num_common = np.count_nonzero(
        ((paths_a == paths_b) & (paths_a >= 0) & (paths_b >= 0)), axis=-1
    )
    return paths[inds_a, num_common - 1]


class FindLCA:
    def __init__(self, tree: Hierarchy):
        self.paths = tree.paths_padded(exclude_root=False)

    def __call__(self, inds_a: np.ndarray, inds_b: np.ndarray) -> np.ndarray:
        paths = self.paths
        paths_a = paths[inds_a]
        paths_b = paths[inds_b]
        num_common = np.count_nonzero(
            ((paths_a == paths_b) & (paths_a >= 0) & (paths_b >= 0)), axis=-1
        )
        return paths[inds_a, num_common - 1]


def truncate_at_lca(tree: Hierarchy, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
    """Truncates the prediction if a descendant of the ground-truth.

    Note that this calls find_lca().
    If calling repetitively, use `FindLCA` and `truncate_given_lca`.
    """
    lca = find_lca(tree, gt, pr)
    return truncate_given_lca(gt, pr, lca)


def truncate_given_lca(gt: np.ndarray, pr: np.ndarray, lca: np.ndarray) -> np.ndarray:
    """Truncates the prediction if a descendant of the ground-truth."""
    return np.where(gt == lca, gt, pr)


def format_tree(
    tree: Hierarchy, node_names: Optional[list[str]] = None, include_size: bool = False
) -> str:
    if node_names is None:
        node_names = [str(i) for i in range(tree.num_nodes())]

    node_to_children = tree.children()
    node_sizes = tree.num_leaf_descendants()

    def subtree(node, node_prefix, desc_prefix):
        name = node_names[node]
        size = node_sizes[node]
        text = f"{name} ({size})" if include_size and size > 1 else name
        yield node_prefix + text + "\n"
        children = node_to_children.get(node, ())
        for i, child in enumerate(children):
            is_last = i == len(children) - 1
            yield from subtree(
                child,
                node_prefix=desc_prefix + ("└── " if is_last else "├── "),
                desc_prefix=desc_prefix + ("    " if is_last else "│   "),
            )

    return "".join(subtree(0, "", ""))


def level_nodes(tree: Hierarchy, extend: bool = False) -> list[np.ndarray]:
    node_depth = tree.depths()
    is_leaf = tree.leaf_mask()
    max_depth = np.max(node_depth)
    level_depth = np.arange(1, max_depth + 1)
    if not extend:
        level_masks = level_depth[:, None] == node_depth
    else:
        level_masks = (level_depth[:, None] == node_depth) | (
            (level_depth[:, None] > node_depth) & is_leaf
        )
    return [np.flatnonzero(mask) for mask in level_masks]


def level_successors(
    tree: Hierarchy,
) -> tuple[list[np.ndarray], list[list[np.ndarray]]]:
    """Returns the list of parents and their children at each depth."""
    node_depth = tree.depths()
    max_depth = np.max(node_depth)
    node_children = tree.children()

    # Get internal nodes at each level.
    level_parents = [[] for _ in range(max_depth)]
    level_children = [[] for _ in range(max_depth)]
    for u in tree.internal_subset():
        d = node_depth[u]
        level_parents[d].append(u)
        level_children[d].append(node_children[u])
    level_sizes = np.array(list(map(len, level_parents)))
    assert np.all(level_sizes > 0)

    level_parents = [np.array(x, dtype=int) for x in level_parents]
    return level_parents, level_children


def level_successors_padded(
    tree: Hierarchy,
    method: str = "constant",
    constant_value: int = -1,
) -> tuple[list[np.ndarray], list[np.ndarray]]:
    """Returns the list of parents and their children at each depth."""
    level_parents, level_children = level_successors(tree)
    level_children = [
        _pad_2d(x, dtype=np.dtype(int), method=method, constant_value=constant_value)
        for x in level_children
    ]
    return level_parents, level_children


def siblings(tree: Hierarchy) -> list[np.ndarray]:
    node_parent = tree.parents()
    node_children = tree.children()
    node_children[-1] = np.empty([0], dtype=int)
    return [_except(u, node_children[node_parent[u]]) for u in range(tree.num_nodes())]


def siblings_padded(
    tree: Hierarchy, method: str = "constant", constant_value: int = -1
) -> np.ndarray:
    ragged = siblings(tree)
    return _pad_2d(
        ragged, dtype=np.dtype(int), method=method, constant_value=constant_value
    )


def _except(v, x: np.ndarray) -> np.ndarray:
    return x[x != v]


def _pad_2d(
    x: list[np.ndarray], dtype: np.dtype, method: str = "constant", constant_value=None
) -> np.ndarray:
    num_rows = len(x)
    row_lens = list(map(len, x))
    num_cols = max(map(len, x))
    if method == "constant":
        assert constant_value is not None
        padded = np.full((num_rows, num_cols), constant_value, dtype=dtype)
    elif method == "first":
        padded = np.tile(
            np.asarray([row[0] for row in x], dtype=dtype)[:, None], num_cols
        )
    elif method == "last":
        padded = np.tile(
            np.asarray([row[-1] for row in x], dtype=dtype)[:, None], num_cols
        )
    elif method == "index":
        padded = np.tile(np.arange(num_rows, dtype=dtype)[:, None], num_cols)
    else:
        raise ValueError("unknown pad method", method)
    row_index = np.concatenate(
        [np.full(row_len, i) for i, row_len in enumerate(row_lens)]
    )
    col_index = np.concatenate([np.arange(row_len) for row_len in row_lens])
    padded[row_index, col_index] = np.concatenate(x)
    return padded


def most_confident_leaf(tree: Hierarchy, p: np.ndarray) -> np.ndarray:
    assert p.shape[-1] == tree.num_nodes()
    is_leaf = tree.leaf_mask()
    return argmax_where(p, is_leaf)


def max_info_majority_subtree(tree: Hierarchy, p: np.ndarray) -> np.ndarray:
    assert p.shape[-1] == tree.num_nodes()
    # -x is a monotonic mapping of -log(1/(x/n)) = log x - constant
    specificity = -tree.num_leaf_descendants()
    # Trivial nodes are allowed in the tree, but we will not predict them.
    # If the parent is equal to the child, it has the same score and specificity.
    # Maybe we should break ties in (specificity, p) using depth?
    not_trivial = tree.num_children() != 1
    return argmax_with_confidence(specificity, p, 0.5, not_trivial)


def pareto_optimal_predictions(
    info: np.ndarray,
    prob: np.ndarray,
    min_threshold: Optional[float] = None,
    condition: Optional[np.ndarray] = None,
    require_unique: bool = False,
) -> np.ndarray:
    """Finds the sequence of nodes that can be chosen by threshold.

    Returns nodes that are more specific than all more-confident predictions.
    This is equivalent to:
    (1) nodes such that there does not exist a node which is more confident and more specific,
    (2) nodes such that all nodes are less confident or less specific (*less than or equal).

    The resulting nodes are ordered descending by prob (and ascending by info).
    """
    assert prob.ndim == 1
    assert info.ndim == 1

    is_valid = np.ones(prob.shape, dtype=bool)
    if min_threshold is not None:
        is_valid = is_valid & (prob > min_threshold)
    if condition is not None:
        is_valid = is_valid & condition
    assert np.any(is_valid), "require at least one valid element"
    prob = prob[is_valid]
    info = info[is_valid]
    (valid_inds,) = np.nonzero(is_valid)

    # Order descending by prob then descending by info.
    # Note that np.lexsort() orders first by the last key.
    # (Performs stable sort from first key to last key.)
    order = np.lexsort((-info, -prob))
    prob = prob[order]
    info = info[order]

    max_info = np.maximum.accumulate(info)
    keep = (prob[1:] > prob[:-1]) | (info[1:] > max_info[:-1])
    keep = np.concatenate(([True], keep))

    if require_unique:
        if np.any(
            (prob[1:] == prob[:-1]) & (info[1:] == info[:-1]) & (keep[1:] | keep[:-1])
        ):
            raise ValueError("set is not unique")

    return valid_inds[order[keep]]


def argmax_where(
    value: np.ndarray, condition: np.ndarray, axis: int = -1, keepdims: bool = False
) -> np.ndarray:
    # Will raise an exception if not np.all(np.any(condition, axis=axis)).
    # return np.nanargmax(np.where(condition, value, np.nan), axis=axis, keepdims=keepdims)
    # nanargmax() only has keepdims for numpy>=1.22
    result = np.nanargmax(np.where(condition, value, np.nan), axis=axis)
    if keepdims:
        result = np.expand_dims(result, axis)
    return result


def max_where(
    value: np.ndarray, condition: np.ndarray, axis: int = -1, keepdims: bool = False
) -> np.ndarray:
    assert np.all(np.any(condition, axis=axis)), "require at least one valid element"
    # return np.nanmax(np.where(condition, value, np.nan), axis=axis, keepdims=keepdims)
    result = np.nanmax(np.where(condition, value, np.nan), axis=axis)
    if keepdims:
        result = np.expand_dims(result, axis)
    return result


def arglexmin(keys: Tuple[np.ndarray, ...], axis: int = -1) -> np.ndarray:
    order = np.lexsort(keys, axis=axis)
    return np.take(order, 0, axis=axis)


def arglexmin_where(
    keys: Tuple[np.ndarray, ...],
    condition: np.ndarray,
    axis: int = -1,
    keepdims: bool = False,
) -> np.ndarray:
    assert np.all(np.any(condition, axis=axis)), "require at least one valid element"
    order = np.lexsort(keys, axis=axis)
    # Take first element in order that satisfies condition.
    first_valid = np.expand_dims(
        np.argmax(np.take_along_axis(condition, order, axis=axis), axis=axis), axis
    )
    result = np.take_along_axis(order, first_valid, axis=axis)
    if not keepdims:
        result = np.squeeze(result, axis=axis)
    return result


def argmax_with_confidence(
    value: np.ndarray,
    p: np.ndarray,
    threshold: float,
    condition: Optional[np.ndarray] = None,
) -> np.ndarray:
    """Finds element that maximizes (value, p) subject to p > threshold."""
    mask = p > threshold
    if condition is not None:
        mask = mask & condition
    return arglexmin_where(np.broadcast_arrays(-p, -value), mask)


def plurality_threshold(
    tree: Hierarchy, p: np.ndarray, axis: int = -1, keepdims: bool = False
) -> np.ndarray:
    assert axis in (-1, p.ndim - 1)
    children = tree.children()
    # Find the top 2 elements of each non-trivial family.
    top2_inds = {
        u: inds[np.argsort(p[..., inds], axis=-1)[..., -2:]]
        for u, inds in children.items()
        if len(inds) > 1
    }
    top2_values = np.stack(
        [np.take_along_axis(p, ind, axis=-1) for ind in top2_inds.values()], axis=-1
    )
    # Take the maximum 2nd-best over all non-trivial families.
    threshold = np.max(top2_values, axis=-1)[..., -2]
    if keepdims:
        threshold = np.expand_dims(threshold, -1)
    return threshold


def _apply_to_maybe(fn: Callable, x: Optional[torch.Tensor]) -> Optional[torch.Tensor]:
    return fn(x) if x is not None else None


class Sum(nn.Module):
    """Implements sum_xxx as an object. Avoids re-computation."""

    def __init__(
        self,
        tree: Hierarchy,
        transpose: bool,
        subset: Optional[np.ndarray] = None,
        # leaf_only: bool = False,
        exclude_root: bool = False,
        strict: bool = False,
    ):
        super().__init__()
        # The value matrix[i, j] is true if i is an ancestor of j.
        # Take transpose for sum over descendants.
        matrix = tree.ancestor_mask(strict=strict)
        if subset is not None:
            matrix = matrix[:, subset]
        if exclude_root:
            matrix = matrix[1:, :]
        if transpose:
            matrix = matrix.T
        matrix = torch.from_numpy(matrix).type(torch.get_default_dtype())
        self.matrix = matrix

    def _apply(self, fn):
        super()._apply(fn)
        self.matrix = fn(self.matrix)
        return self

    def forward(self, values: torch.Tensor, dim: int = -1) -> torch.Tensor:
        assert dim in (-1, values.ndim - 1)
        return torch.tensordot(values, self.matrix, dims=1)


SumAncestors = partial(Sum, transpose=False)
SumDescendants = partial(Sum, transpose=True)


class HierSoftmaxCrossEntropy(nn.Module):
    """Implements cross-entropy for YOLO-style conditional softmax. Avoids re-computation.

    Supports integer label targets or distribution targets.
    """

    def __init__(
        self,
        tree: Hierarchy,
        with_leaf_targets: bool = False,
        label_smoothing: float = 0.0,
        node_weight: Optional[torch.Tensor] = None,
    ):
        super().__init__()
        self.label_smoothing = label_smoothing
        if with_leaf_targets:
            self.label_order = torch.from_numpy(tree.leaf_subset())
            self.num_labels = len(self.label_order)
        else:
            self.label_order = None
            self.num_labels = tree.num_nodes()
        self.hier_cond_log_softmax = HierCondLogSoftmax(tree)
        self.sum_label_descendants = SumDescendants(tree, subset=self.label_order)
        self.prior = torch.from_numpy(uniform_leaf(tree))
        self.node_weight = node_weight

    def _apply(self, fn):
        super()._apply(fn)
        self.label_order = _apply_to_maybe(fn, self.label_order)
        self.hier_cond_log_softmax = self.hier_cond_log_softmax._apply(fn)
        self.sum_label_descendants = self.sum_label_descendants._apply(fn)
        self.prior = fn(self.prior)
        self.node_weight = _apply_to_maybe(fn, self.node_weight)
        return self

    def forward(
        self, scores: torch.Tensor, labels: torch.Tensor, dim: int = -1
    ) -> torch.Tensor:
        assert labels.ndim in [scores.ndim, scores.ndim - 1]
        assert dim in (-1, scores.ndim - 1)
        # Convert labels to one-hot if they are not.
        if labels.ndim < scores.ndim:
            labels = F.one_hot(labels, self.num_labels)
        labels = labels.type(torch.get_default_dtype())
        q = self.sum_label_descendants(labels)
        if self.label_smoothing:
            q = (1 - self.label_smoothing) * q + self.label_smoothing * self.prior
        log_cond_p = self.hier_cond_log_softmax(scores, dim=-1)
        xent = q * -log_cond_p
        if self.node_weight is not None:
            xent = xent * self.node_weight
        xent = torch.sum(xent, dim=-1)
        return torch.mean(xent)


def hier_cond_log_softmax(
    tree: Hierarchy, scores: torch.Tensor, dim: int = -1
) -> torch.Tensor:
    """Returns log-likelihood of each node given its parent."""
    assert dim == -1 or dim == scores.ndim - 1
    num_nodes = tree.num_nodes()
    num_internal = tree.num_internal_nodes()
    node_to_children = tree.children()
    cond_children = [node_to_children[x] for x in tree.internal_subset()]
    cond_num_children = list(map(len, cond_children))
    max_num_children = max(cond_num_children)
    row_index = np.concatenate([np.full(n, i) for i, n in enumerate(cond_num_children)])
    col_index = np.concatenate([np.arange(n) for n in cond_num_children])
    flat_index = row_index * max_num_children + col_index
    child_index = np.concatenate(cond_children)

    device = scores.device
    flat_index = torch.from_numpy(flat_index).to(device)
    child_index = torch.from_numpy(child_index).to(device)
    input_shape = list(scores.shape)
    flat_shape = [*input_shape[:-1], num_internal * max_num_children]
    # Pad with -inf for log_softmax.
    # flat[..., flat_index] = scores
    flat = torch.full(flat_shape, -torch.inf, device=device).index_copy(
        -1, flat_index, scores
    )
    split_shape = [*input_shape[:-1], num_internal, max_num_children]
    child_scores = flat.reshape(split_shape)
    child_log_p = F.log_softmax(child_scores, dim=-1)
    child_log_p = child_log_p.reshape(flat_shape)
    output_shape = [*input_shape[:-1], num_nodes]
    # log_cond_p[..., child_index] = child_log_p[..., flat_index]
    log_cond_p = torch.zeros(output_shape, device=device).index_copy(
        -1, child_index, child_log_p.index_select(-1, flat_index)
    )
    return log_cond_p


def hier_log_softmax(
    tree: Hierarchy, scores: torch.Tensor, dim: int = -1
) -> torch.Tensor:
    """Returns log-likelihood for conditional softmax."""
    # Finally, take sum over ancestor conditionals to obtain likelihoods.
    assert dim in (-1, scores.ndim - 1)
    log_cond_p = hier_cond_log_softmax(tree, scores, dim=dim)
    device = scores.device
    sum_ancestors_fn = SumAncestors(tree, exclude_root=True).to(device)
    return sum_ancestors_fn(log_cond_p, dim=-1)


class HierCondLogSoftmax(nn.Module):
    """Implements hier_cond_log_softmax as an object. Avoids re-computation."""

    def __init__(self, tree: Hierarchy):
        super().__init__()
        num_nodes = tree.num_nodes()
        num_internal = tree.num_internal_nodes()
        node_to_children = tree.children()
        cond_children = [node_to_children[x] for x in tree.internal_subset()]
        cond_num_children = list(map(len, cond_children))
        max_num_children = max(cond_num_children)
        row_index = np.concatenate(
            [np.full(n, i) for i, n in enumerate(cond_num_children)]
        )
        col_index = np.concatenate([np.arange(n) for n in cond_num_children])
        flat_index = torch.from_numpy(row_index * max_num_children + col_index)
        child_index = torch.from_numpy(np.concatenate(cond_children))

        self.num_nodes = num_nodes
        self.num_internal = num_internal
        self.max_num_children = max_num_children
        self.flat_index = flat_index
        self.child_index = child_index

    def _apply(self, fn):
        super()._apply(fn)
        self.flat_index = fn(self.flat_index)
        self.child_index = fn(self.child_index)
        return self

    def forward(self, scores: torch.Tensor, dim: int = -1) -> torch.Tensor:
        assert dim in (-1, scores.ndim - 1)
        device = scores.device
        input_shape = list(scores.shape)
        flat_shape = [*input_shape[:-1], self.num_internal * self.max_num_children]
        # Pad with -inf for log_softmax.
        # flat[..., flat_index] = scores
        flat = torch.full(flat_shape, -torch.inf, device=device).index_copy(
            -1, self.flat_index, scores
        )
        split_shape = [*input_shape[:-1], self.num_internal, self.max_num_children]
        child_scores = flat.reshape(split_shape)
        child_log_p = F.log_softmax(child_scores, dim=-1)
        child_log_p = child_log_p.reshape(flat_shape)
        output_shape = [*input_shape[:-1], self.num_nodes]
        # log_cond_p[..., child_index] = child_log_p[..., flat_index]
        log_cond_p = torch.zeros(output_shape, device=device).index_copy(
            -1, self.child_index, child_log_p.index_select(-1, self.flat_index)
        )
        return log_cond_p


class HierLogSoftmax(nn.Module):
    """Implements hier_log_softmax as an object. Avoids re-computation."""

    def __init__(self, tree: Hierarchy):
        super().__init__()
        self.cond_log_softmax = HierCondLogSoftmax(tree)
        self.sum_ancestors_fn = SumAncestors(tree, exclude_root=False)

    def _apply(self, fn):
        super()._apply(fn)
        self.cond_log_softmax = self.cond_log_softmax._apply(fn)
        self.sum_ancestors_fn = self.sum_ancestors_fn._apply(fn)
        return self

    def forward(self, scores: torch.Tensor, dim: int = -1) -> torch.Tensor:
        log_cond_p = self.cond_log_softmax(scores, dim=dim)
        return self.sum_ancestors_fn(log_cond_p, dim=dim)


def multilabel_log_likelihood(
    scores: torch.Tensor,
    dim: int = -1,
    insert_root: bool = False,
    replace_root: bool = False,
    temperature: Optional[float] = None,
) -> torch.Tensor:
    assert not (insert_root and replace_root)
    assert dim in (-1, scores.ndim - 1)
    device = scores.device
    if temperature:
        scores = scores / temperature
    logp = F.logsigmoid(scores)
    if insert_root:
        zero = torch.zeros((*scores.shape[:-1], 1), device=device)
        logp = torch.cat([zero, logp], dim=-1)
    elif replace_root:
        zero = torch.zeros((*scores.shape[:-1], 1), device=device)
        logp = logp.index_copy(-1, torch.tensor([0], device=device), zero)
    return logp


class RandomCut(nn.Module):
    """Samples a random cut of the tree.

    Starts from the root and samples from Bernoulli for each node.
    If a node's sample is 0, the tree is terminated at that node.

    Returns a binary mask for the leaf nodes of the cut.
    """

    def __init__(self, tree: Hierarchy, cut_prob: float, permit_root_cut: bool = False):
        super().__init__()
        self.num_nodes = tree.num_nodes()
        self.cut_prob = cut_prob
        self.permit_root_cut = permit_root_cut
        self.sum_ancestors = SumAncestors(tree)
        self.node_parent = torch.from_numpy(tree.parents(root_loop=True))

    def _apply(self, fn):
        super()._apply(fn)
        self.sum_ancestors._apply(fn)
        self.node_parent = fn(self.node_parent)
        return self

    def forward(self, batch_shape: Sequence[int]) -> torch.Tensor:
        device = self.node_parent.device
        # Random Bernoulli for every node.
        drop = torch.bernoulli(
            torch.full((*batch_shape, self.num_nodes), self.cut_prob)
        )
        drop = drop.to(device=device)
        if not self.permit_root_cut:
            drop[..., 0] = 0
        # Check whether to keep all ancestors of each node (drop zero ancestors).
        subtree_mask = self.sum_ancestors(drop, dim=-1) == 0
        # Dilate to keep nodes whose parents belong to subtree.
        subtree_mask = subtree_mask[..., self.node_parent]
        subtree_mask[..., 0] = True  # Root is always kept.
        # Count number of children (do not count parent as child of itself).
        num_children = torch.zeros((*batch_shape, self.num_nodes), device=device)
        num_children = num_children.index_add(
            -1, self.node_parent[1:], subtree_mask[..., 1:].float()
        )
        # Find boundary of subtree.
        boundary = subtree_mask & (num_children == 0)
        return boundary


class RandomCutLoss(nn.Module):
    """Cross-entropy loss using the leaf nodes of a random cut.

    As described in "Deep RTC" (Wu et al., 2020).
    """

    def __init__(
        self,
        tree: Hierarchy,
        cut_prob: float,
        permit_root_cut: bool = False,
        with_leaf_targets: bool = True,
    ):
        super().__init__()
        is_ancestor = tree.ancestor_mask(strict=False)
        # label_to_targets[gt, pr] = 1 iff pr `is_ancestor` gt
        label_to_targets = is_ancestor.T
        if with_leaf_targets:
            label_order = tree.leaf_subset()
            label_to_targets = label_to_targets[label_order]
        else:
            # Need to use FlatSoftmaxNLL?
            raise NotImplementedError

        self.random_cut_fn = RandomCut(
            tree, cut_prob=cut_prob, permit_root_cut=permit_root_cut
        )
        self.label_to_targets = torch.from_numpy(label_to_targets)

    def _apply(self, fn):
        super()._apply(fn)
        self.random_cut_fn._apply(fn)
        self.label_to_targets = fn(self.label_to_targets)
        return self

    def forward(self, scores: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        labels = torch.argmax(labels, dim=1)
        cut = self.random_cut_fn(scores.shape[:-1])
        targets = self.label_to_targets[labels.long(), :]
        # No loss for root node.
        cut = cut[..., 1:]
        targets = targets[..., 1:]
        # Obtain targets in cut subset.
        cut_targets = cut & targets
        assert torch.all(torch.sum(cut_targets, dim=-1) == 1)
        # loss = F.cross_entropy(cut_scores, cut_targets, reduction='none')
        neg_inf = torch.tensor(-torch.inf, device=scores.device)
        zero = torch.tensor(0.0, device=scores.device)
        pos_score = torch.sum(torch.where(cut_targets, scores, zero), dim=-1)
        loss = -pos_score + torch.logsumexp(torch.where(cut, scores, neg_inf), dim=-1)
        return torch.mean(loss)


class LCAMetric:
    def __init__(self, tree: Hierarchy, value: np.ndarray):
        self.value = value
        self.find_lca = FindLCA(tree)

    def value_at_lca(self, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
        lca = self.find_lca(gt, pr)
        return self.value[lca]

    def value_at_gt(self, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
        gt, _ = np.broadcast_arrays(gt, pr)
        return self.value[gt]

    def value_at_pr(self, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
        _, pr = np.broadcast_arrays(gt, pr)
        return self.value[pr]

    def deficient(self, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
        lca = self.find_lca(gt, pr)
        return self.value[gt] - self.value[lca]

    def excess(self, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
        lca = self.find_lca(gt, pr)
        return self.value[pr] - self.value[lca]

    def dist(self, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
        lca = self.find_lca(gt, pr)
        excess = self.value[pr] - self.value[lca]
        deficient = self.value[gt] - self.value[lca]
        return excess + deficient

    def recall(self, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
        lca = self.find_lca(gt, pr)
        gt_value = self.value[gt]
        lca_value = self.value[lca]
        with np.errstate(invalid="ignore"):
            return np.where(
                (lca_value == 0) & (gt_value == 0), 1.0, lca_value / gt_value
            )

    def precision(self, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
        lca = self.find_lca(gt, pr)
        pr_value = self.value[pr]
        lca_value = self.value[lca]
        with np.errstate(invalid="ignore"):
            return np.where(
                (lca_value == 0) & (pr_value == 0), 1.0, lca_value / pr_value
            )

    def f1(self, gt: np.ndarray, pr: np.ndarray) -> np.ndarray:
        lca = self.find_lca(gt, pr)
        gt_value = self.value[gt]
        pr_value = self.value[pr]
        lca_value = self.value[lca]
        with np.errstate(invalid="ignore"):
            r = np.where((lca_value == 0) & (gt_value == 0), 1.0, lca_value / gt_value)
            p = np.where((lca_value == 0) & (pr_value == 0), 1.0, lca_value / pr_value)
        with np.errstate(divide="ignore"):
            return 2 / (1 / r + 1 / p)


class MarginLoss(nn.Module):
    """Computes soft or hard margin loss for given a margin function."""

    def __init__(
        self,
        tree: Hierarchy,
        with_leaf_targets: bool,
        hardness: str = "soft",
        margin: str = "depth_dist",
        tau: float = 1.0,
    ):
        super().__init__()
        if hardness not in ("soft", "hard"):
            raise ValueError("unknown hardness", hardness)
        n = tree.num_nodes()
        label_order = tree.leaf_subset() if with_leaf_targets else np.arange(n)

        # Construct array label_margin[gt_label, pr_node].
        if margin in ("edge_dist", "depth_dist"):
            # label_margin = metrics.edge_dist(tree, label_order[:, None], np.arange(n)[None, :])
            depth = tree.depths()
            margin_arr = LCAMetric(tree, depth).dist(label_order[:, None], np.arange(n))
        elif margin == "incorrect":
            is_ancestor = tree.ancestor_mask()
            is_correct = is_ancestor[:, label_order].T
            margin_arr = 1 - is_correct
        elif margin == "info_dist":
            info = np.log(tree.num_leaf_nodes() / tree.num_leaf_descendants())
            margin_arr = LCAMetric(tree, info).dist(label_order[:, None], np.arange(n))
        elif margin == "depth_deficient":
            depth = tree.depths()
            margin_arr = LCAMetric(tree, depth).deficient(
                label_order[:, None], np.arange(n)
            )
        elif margin == "log_depth_f1_error":
            depth = tree.depths()
            margin_arr = np.log(
                1 - LCAMetric(tree, depth).f1(label_order[:, None], np.arange(n))
            )
        else:
            raise ValueError("unknown margin", margin)

        self.hardness = hardness
        self.tau = tau
        self.label_order = torch.from_numpy(label_order)
        self.margin = torch.from_numpy(margin_arr)

    def _apply(self, fn):
        super()._apply(fn)
        self.label_order = fn(self.label_order)
        self.margin = fn(self.margin)
        return self

    def forward(self, scores: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        labels = torch.argmax(labels, dim=1)
        label_node = (
            labels if self.label_order is None else self.label_order[labels.long()]
        )
        label_score = scores.gather(-1, label_node.unsqueeze(-1)).squeeze(-1)
        label_margin = self.margin[labels.long(), :]
        if self.hardness == "soft":
            loss = -label_score + torch.logsumexp(
                scores + self.tau * label_margin, dim=-1
            )
        elif self.hardness == "hard":
            loss = torch.relu(
                torch.max(
                    scores - label_score.unsqueeze(-1) + self.tau * label_margin,
                    dim=-1,
                )[0]
            )
        else:
            assert False
        return torch.mean(loss)


class FlatSoftmaxNLL(nn.Module):
    """Like cross_entropy() but supports internal labels."""

    def __init__(self, tree, with_leaf_targets: bool = False, reduction: str = "mean"):
        super().__init__()
        assert reduction in ("mean", "none", None)
        if with_leaf_targets:
            raise ValueError("use F.cross_entropy() instead!")
        # The value is_ancestor[i, j] is true if node i is an ancestor of node j.
        is_ancestor = tree.ancestor_mask(strict=False)
        leaf_masks = is_ancestor[:, tree.leaf_mask()]
        self.leaf_masks = torch.from_numpy(leaf_masks)
        self.reduction = reduction

    def _apply(self, fn):
        super()._apply(fn)
        self.leaf_masks = fn(self.leaf_masks)
        return self

    def forward(self, scores: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        labels = torch.argmax(labels, dim=1)
        logp_leaf = F.log_softmax(scores, dim=-1)
        # Obtain logp for leaf descendants, -inf for other nodes.
        label_leaf_mask = self.leaf_masks[labels.long(), :]
        inf = torch.tensor(torch.inf, device=scores.device)
        logp_ancestors = torch.where(label_leaf_mask, logp_leaf, -inf)
        logp_label = torch.logsumexp(logp_ancestors, dim=-1)
        loss = -logp_label
        if self.reduction == "mean":
            return torch.mean(loss)
        else:
            return loss


def SumLeafDescendants(tree: Hierarchy, **kwargs):
    return SumDescendants(tree, subset=tree.leaf_mask(), **kwargs)



import unittest
import vamb
import testtools
from pathlib import Path
import tempfile
import shutil
import io


class TestParseMarkers(unittest.TestCase):
    def test_instantiate(self):
        tmp = tempfile.mkdtemp()
        tmp_path = Path(tmp)
        shutil.rmtree(tmp)
        markers = vamb.parsemarkers.Markers.from_files(
            Path(testtools.DATADIR).joinpath("marker.fna"),
            Path(testtools.PARENTDIR).joinpath("vamb").joinpath("marker.hmm"),
            ["abc"],
            tmp_path,
            4,
            None,
        )
        self.assertIsNotNone(markers.markers[0])
        self.assertEqual(len(markers.markers), 1)
        self.assertEqual(set(markers.markers[0]), {39})
        self.assertEqual(
            markers.refhash, vamb.vambtools.RefHasher.hash_refnames(["abc"])
        )

        buf = io.StringIO()
        markers.save(buf)
        buf.seek(0)

        markers2 = vamb.parsemarkers.Markers.load(buf, markers.refhash)
        self.assertEqual(len(markers.markers), len(markers2.markers))
        self.assertEqual(set(markers.markers[0]), set(markers2.markers[0]))
        self.assertEqual(markers.marker_names, markers2.marker_names)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
from importlib import metadata


# -- Project information -----------------------------------------------------

project = "vamb"
copyright = "2024, Jakob Nybo Nissen, Simon Rasmussen"  # ! please update
author = "Jakob Nybo Nissen, Simon Rasmussen"
PACKAGE_VERSION = metadata.version("vamb")
version = PACKAGE_VERSION
release = PACKAGE_VERSION


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autodoc.typehints",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx_new_tab_link",
    "myst_nb",
]

#  https://myst-nb.readthedocs.io/en/latest/computation/execute.html
nb_execution_mode = "auto"

myst_enable_extensions = ["dollarmath", "amsmath"]

# Plolty support through require javascript library
# https://myst-nb.readthedocs.io/en/latest/render/interactive.html#plotly
html_js_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js"
]

# https://myst-nb.readthedocs.io/en/latest/configuration.html
# Execution
nb_execution_raise_on_error = True
# Rendering
nb_merge_streams = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    ".npz",
]


# Intersphinx options
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "torch": ("https://pytorch.org/docs/stable/index.html", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    # "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    # "scikit-learn": ("https://scikit-learn.org/stable/", None),
    # "matplotlib": ("https://matplotlib.org/stable/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# See:
# https://github.com/executablebooks/MyST-NB/blob/master/docs/conf.py
# html_title = ""
html_theme = "sphinx_book_theme"
# html_logo = "_static/logo-wide.svg"
# html_favicon = "_static/logo-square.svg"
html_theme_options = {
    "github_url": "https://github.com/RasmussenLab/vamb",
    "repository_url": "https://github.com/RasmussenLab/vamb",
    "repository_branch": "main",
    "home_page_in_toc": True,
    "path_to_docs": "docs",
    "show_navbar_depth": 2,
    "use_edit_page_button": True,
    "use_repository_button": True,
    "use_download_button": True,
    "launch_buttons": {
        "colab_url": "https://colab.research.google.com"
        #     "binderhub_url": "https://mybinder.org",
        #     "notebook_interface": "jupyterlab",
    },
    "navigation_with_keys": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]


# -- Setup for sphinx-apidoc -------------------------------------------------

# Read the Docs doesn't support running arbitrary commands like tox.
# sphinx-apidoc needs to be called manually if Sphinx is running there.
# https://github.com/readthedocs/readthedocs.org/issues/1139

if os.environ.get("READTHEDOCS") == "True":
    from pathlib import Path

    PROJECT_ROOT = Path(__file__).parent.parent
    PACKAGE_ROOT = PROJECT_ROOT / "vamb"

#     def run_apidoc(_):
#         from sphinx.ext import apidoc
#
#         apidoc.main(
#             [
#                 "--force",
#                 "--implicit-namespaces",
#                 "--module-first",
#                 "--separate",
#                 "-o",
#                 str(PROJECT_ROOT / "doc" / "reference"),
#                 str(PACKAGE_ROOT),
#                 str(PACKAGE_ROOT / "*.c"),
#                 str(PACKAGE_ROOT / "*.so"),
#             ]
#         )
#
#     def setup(app):
#         app.connect("builder-inited", run_apidoc)



"""
The following code is based on the k-means based reclustering algorithm first published at https://github.com/BigDataBiology/SemiBin
The original code is distributed under MIT License.
"""

from sklearn.cluster import KMeans
import numpy as np
from collections import defaultdict
from sklearn.cluster import DBSCAN
from sklearn.metrics import pairwise_distances
from vamb.taxonomy import Taxonomy
from vamb.parsemarkers import Markers, MarkerID
from vamb.parsecontigs import CompositionMetaData
from vamb.vambtools import RefHasher
from collections.abc import Sequence, Iterable
from typing import NewType, Optional, Union

# We use these aliases to be able to work with integers, which is faster.
ContigId = NewType("ContigId", int)
BinId = NewType("BinId", int)

# TODO: We might want to benchmark the best value for this constant.
# Right now, we do too much duplicated work by clustering 18 times.
EPS_VALUES = np.arange(0.01, 0.35, 0.02)


class KmeansAlgorithm:
    "Arguments needed specifically when using the KMeans algorithm"

    def __init__(
        self, clusters: list[set[ContigId]], random_seed: int, contiglengths: np.ndarray
    ):
        assert np.issubdtype(contiglengths.dtype, np.integer)
        self.contiglengths = contiglengths
        self.clusters = clusters
        self.random_seed = random_seed


class DBScanAlgorithm:
    "Arguments needed specifically when using the DBScan algorithm"

    def __init__(
        self, comp_metadata: CompositionMetaData, taxonomy: Taxonomy, n_processes: int
    ):
        if not taxonomy.is_canonical:
            raise ValueError(
                "Can only run DBScan on a Taxonomy object with is_canonical set"
            )
        RefHasher.verify_refhash(
            taxonomy.refhash,
            comp_metadata.refhash,
            "taxonomy",
            "composition",
            None,
        )
        self.contiglengths = comp_metadata.lengths
        self.taxonomy = taxonomy
        self.n_processes = n_processes


def recluster_bins(
    markers: Markers,
    latent: np.ndarray,
    algorithm: Union[KmeansAlgorithm, DBScanAlgorithm],
) -> list[set[ContigId]]:
    assert np.issubdtype(algorithm.contiglengths.dtype, np.integer)
    assert np.issubdtype(latent.dtype, np.floating)

    if not (len(algorithm.contiglengths) == markers.n_seqs == len(latent)):
        raise ValueError(
            "Number of elements in contiglengths, markers and latent must match"
        )

    # Simply dispatch to the right implementation based on the algorithm used
    if isinstance(algorithm, KmeansAlgorithm):
        return recluster_kmeans(
            algorithm.clusters,
            latent,
            algorithm.contiglengths,
            markers,
            algorithm.random_seed,
        )
    elif isinstance(algorithm, DBScanAlgorithm):
        assert len(algorithm.taxonomy.contig_taxonomies) == markers.n_seqs
        return recluster_dbscan(
            algorithm.taxonomy,
            latent,
            algorithm.contiglengths,
            markers,
            algorithm.n_processes,
        )


def recluster_kmeans(
    clusters: list[set[ContigId]],
    latent: np.ndarray,
    contiglengths: np.ndarray,
    markers: Markers,
    random_seed: int,
) -> list[set[ContigId]]:
    assert len(latent) == len(contiglengths) == markers.n_seqs
    assert np.issubdtype(contiglengths.dtype, np.integer)
    assert np.issubdtype(latent.dtype, np.floating)
    assert latent.ndim == 2

    result: list[set[ContigId]] = []
    indices_by_medoid: dict[int, set[ContigId]] = defaultdict(set)
    # We loop over all existing clusters, and determine if they should be split,
    # by looking at the median number of single-copy genes in the cluster
    for cluster in clusters:
        # All clusters with 1 contig by definition cannot have multiple single-copy
        # genes (because SCGs are deduplicated within a single contig)
        if len(cluster) == 1:
            result.append(cluster)
            continue
        # Get a count of each marker and compute the median count of SCGs
        counts = count_markers(cluster, markers)
        cp = counts.copy()
        cp.sort()
        median_counts: int = cp[len(cp) // 2]
        # If we have less than 2 SCGs on average, the cluster should not be split,
        # and we emit it unchanged
        if median_counts < 2:
            result.append(cluster)
            continue

        # Run K-means with the median number of SCGs to split the contig.
        # We weigh the contigs by length.
        seeds = get_kmeans_seeds(
            cluster,
            markers,
            contiglengths,  # type: ignore
            counts,
            median_counts,
        )

        cluster_indices = np.array(list(cluster))
        cluter_latent = latent[cluster_indices]
        cluster_lengths = contiglengths[cluster_indices]
        seed_latent = latent[seeds]
        kmeans = KMeans(
            n_clusters=median_counts,
            init=seed_latent,
            n_init=1,
            random_state=random_seed,
        )
        kmeans.fit(cluter_latent, sample_weight=cluster_lengths)
        indices_by_medoid.clear()
        for cluster_label, index in zip(kmeans.labels_, cluster_indices):
            indices_by_medoid[cluster_label].add(ContigId(index))
        result.extend(indices_by_medoid.values())

    return result


# Get a vector of counts, of each SCG,
# where if MarkerID(5) is seen 9 times, then counts[5] == 9.
def count_markers(
    contigs: Iterable[ContigId],
    markers: Markers,
) -> np.ndarray:
    counts = np.zeros(markers.n_markers, dtype=np.int32)
    for contig in contigs:
        m = markers.markers[contig]
        if m is not None:
            counts[m] += 1
    return counts


# Same as above, but once we see a very high number of marker genes,
# we bail. This is because a large fraction of time spent in this module
# would otherwise be counting markers of huge clusters, long after we already
# know it's hopelessly contaminated
def count_markers_saturated(
    contigs: Iterable[ContigId],
    markers: Markers,
) -> Optional[np.ndarray]:
    counts = np.zeros(markers.n_markers, dtype=np.int32)
    n_markers = 0
    n_unique = 0
    # This implies contamination == 1.0
    max_duplicates = 1 * markers.n_markers
    for contig in contigs:
        m = markers.markers[contig]
        if m is not None:
            n_markers += len(m)
            for i in m:
                existing = counts[i]
                n_unique += existing == 0
                counts[i] = existing + 1

            if (n_markers - n_unique) > max_duplicates:
                return None
    return counts


# This is not very effectively implemented, but I assume it does not matter.
# This function looks at all markers that occur exactly `median` times, each of these
# markers corresponding to a list of `median` number of contigs.
# It picks the marker for which the smallest contig that contains it is largest.
# The idea here is that long contigs, which contain one of the SCGs that exist exactly
# `median` times are most likely to be close to the actual medoid
# that Kmeans needs to find.
# This is just one possible seeding strategy. We could also plausibly choose e.g.
# the trio of contigs that have the most SCGs.
def get_kmeans_seeds(
    contigs: Iterable[ContigId],
    markers: Markers,
    contiglengths: Sequence[int],
    counts: np.ndarray,
    median: int,
) -> list[ContigId]:
    considered_markers = {MarkerID(i) for (i, c) in enumerate(counts) if c == median}
    contigs_of_markers: dict[MarkerID, list[ContigId]] = defaultdict(list)
    for contig in contigs:
        m = markers.markers[contig]
        if m is None:
            continue
        for mid in m:
            if mid not in considered_markers:
                continue
            contigs_of_markers[MarkerID(mid)].append(contig)

    candidate_list = list(contigs_of_markers.items())
    pair = max(candidate_list, key=lambda x: min(contiglengths[i] for i in x[1]))
    result = pair[1]
    assert len(result) == median
    return result


def get_completeness_contamination(counts: np.ndarray) -> tuple[float, float]:
    n_total = counts.sum()
    n_unique = (counts > 0).sum()
    completeness = n_unique / len(counts)
    contamination = (n_total - n_unique) / len(counts)
    return (completeness, contamination)


def recluster_dbscan(
    taxonomy: Taxonomy,
    latent: np.ndarray,
    contiglengths: np.ndarray,
    markers: Markers,
    num_processes: int,
) -> list[set[ContigId]]:
    # Since DBScan is computationally expensive, and scales poorly with the number
    # of contigs, we use taxonomy to only cluster within each genus
    n_worse_in_row = 0
    genera_indices = group_indices_by_genus(taxonomy)
    best_score = 0
    best_bins: list[set[ContigId]] = []
    for eps in EPS_VALUES:
        bins: list[set[ContigId]] = []
        for indices in genera_indices:
            genus_clusters = dbscan_genus(
                latent[indices], indices, contiglengths[indices], num_processes, eps
            )
            bins.extend(genus_clusters)

        score = count_good_genomes(bins, markers)
        if best_score == 0 or score > best_score:
            best_bins = bins
            best_score = score

        if score >= best_score:
            n_worse_in_row = 0
        else:
            n_worse_in_row += 1
            if n_worse_in_row > 2:
                break

    return best_bins


# DBScan within the subset of contigs that are annotated with a single genus
def dbscan_genus(
    latent_of_genus: np.ndarray,
    original_indices: np.ndarray,
    contiglengths_of_genus: np.ndarray,
    num_processes: int,
    eps: float,
) -> list[set[ContigId]]:
    assert len(latent_of_genus) == len(original_indices) == len(contiglengths_of_genus)
    # Precompute distance matrix. This is O(N^2), but DBScan is even worse,
    # so this pays off.
    # TODO: Maybe we should emit a warning if this function is called with too
    # many points such that this matrix becomes huge?
    distance_matrix = pairwise_distances(
        latent_of_genus, latent_of_genus, metric="cosine"
    )
    # The DBScan approach works by blindly clustering with different eps values
    # (a critical parameter for DBscan), and then using SCGs to select the best
    # subset of clusters.
    # It's ugly and wasteful, but it does work.
    dbscan = DBSCAN(
        eps=eps,
        min_samples=5,
        n_jobs=num_processes,
        metric="precomputed",
    )
    dbscan.fit(distance_matrix, sample_weight=contiglengths_of_genus)
    bins: dict[int, set[ContigId]] = defaultdict(set)
    for original_index, bin_index in zip(original_indices, dbscan.labels_):
        bins[bin_index].add(ContigId(original_index))
    return list(bins.values())


def count_good_genomes(binning: Iterable[Iterable[ContigId]], markers: Markers) -> int:
    max_contamination = 0.3
    min_completeness = 0.75
    result = 0
    for contigs in binning:
        count = count_markers_saturated(contigs, markers)
        if count is None:
            continue
        (comp, cont) = get_completeness_contamination(count)
        if comp >= min_completeness and cont <= max_contamination:
            result += 1

    return result


def group_indices_by_genus(
    taxonomy: Taxonomy,
) -> list[np.ndarray]:
    if not taxonomy.is_canonical:
        raise ValueError("Can only group by genus for a canonical taxonomy")
    by_genus: dict[Optional[str], list[ContigId]] = defaultdict(list)
    for i, tax in enumerate(taxonomy.contig_taxonomies):
        genus = None if tax is None else tax.genus
        by_genus[genus].append(ContigId(i))
    return [np.array(i, dtype=np.int32) for i in by_genus.values()]



__doc__ = "Iterative medoid clustering"

import random as _random
import numpy as _np
import torch as _torch
from collections import deque as _deque, OrderedDict
from math import ceil as _ceil
from torch.functional import Tensor as _Tensor
import vamb.vambtools as _vambtools
from typing import TypeVar, Union, Optional, cast

_DEFAULT_RADIUS = 0.06
# Distance within which to search for medoid point
_MEDOID_RADIUS = 0.05

_DELTA_X = 0.005
_XMAX = 0.3

# We don't want to keep an indefinite amount of data points in the medoid cache,
# since it can be a memory hog.
MAX_CACHED_RESULTS = 64

Cl = TypeVar("Cl", bound="Cluster")


class Loner:
    __slots__ = []
    pass


class NoThreshold:
    __slots__ = []
    pass


# This is the PDF of normal with µ=0, s=0.01 from -0.075 to 0.075 with intervals
# of DELTA_X, for a total of 31 values. We multiply by _DELTA_X so the density
# of one point sums to approximately one
_NORMALPDF = _DELTA_X * _Tensor(
    [
        2.43432053e-11,
        9.13472041e-10,
        2.66955661e-08,
        6.07588285e-07,
        1.07697600e-05,
        1.48671951e-04,
        1.59837411e-03,
        1.33830226e-02,
        8.72682695e-02,
        4.43184841e-01,
        1.75283005e00,
        5.39909665e00,
        1.29517596e01,
        2.41970725e01,
        3.52065327e01,
        3.98942280e01,
        3.52065327e01,
        2.41970725e01,
        1.29517596e01,
        5.39909665e00,
        1.75283005e00,
        4.43184841e-01,
        8.72682695e-02,
        1.33830226e-02,
        1.59837411e-03,
        1.48671951e-04,
        1.07697600e-05,
        6.07588285e-07,
        2.66955661e-08,
        9.13472041e-10,
        2.43432053e-11,
    ]
)


class Cluster:
    __slots__ = [
        "medoid",
        "seed",
        "members",
        "maximal_pvr",
        "observed_pvr",
        "radius",
        "isdefault",
        "successes",
        "attempts",
    ]

    def __init__(
        self,
        medoid: int,
        seed: int,
        members: _np.ndarray,
        maximal_pvr: float,
        # If loner or default: None
        observed_pvr: Optional[float],
        # If loner: None. If default, _DEFAULT_RADIUS
        radius: Optional[float],
        successes: int,
        attempts: int,
    ):
        self.medoid = medoid
        self.seed = seed
        self.members = members
        self.maximal_pvr = maximal_pvr
        self.observed_pvr = observed_pvr
        self.radius = radius
        self.successes = successes
        self.attempts = attempts

    @property
    def kind_str(self) -> str:
        if self.observed_pvr is not None:
            return "normal"
        else:
            if self.radius is None:
                return "loner"
            else:
                return "fallback"


class ClusterGenerator:
    """Iterative medoid cluster generator. Iterate this object to get clusters.

    Inputs:
        matrix: A (obs x features) Numpy matrix of data type numpy.float32
        maxsteps: Stop searching for optimal medoid after N futile attempts [25]
        windowsize: Length of window to count successes [200]
        minsuccesses: Minimum acceptable number of successes [15]
        destroy: Save memory by destroying matrix while clustering [False]
        normalized: Matrix is already preprocessed [False]
        cuda: Accelerate clustering with GPU [False]
    """

    __slots__ = [
        # Maximum number of futile steps taken in the wander_medoid function until it gives up
        # and emits the currently best medoid as the medoid
        "maxsteps",
        # Minimum number of successul clusters emitted of the last `attempts`, lest the
        # peak_valley_ratio be increased to prevent an infinite loop
        "minsuccesses",
        # Whether this clusterer runs on GPU
        "cuda",
        # Random number generator, currently used in wander_medoid function
        "rng",
        # Actual data to be clustered
        "matrix",
        # Lengths of input contigs
        "lengths",
        # This are the original indices of the rows of the matrix. Initially this is just 1..len(matrix),
        # but if not on GPU, we delete used rows of the matrix (and indices) to speed up subsequent computations.
        # Then, we can obtain the original row index by looking up in this array
        "indices",
        # Original indices are scored from most promising as seeds to worse, and ordered in the `order` array.
        # The best contig indices are first.
        "order",
        # The integer index in `order` which the clusterer on next iteration will try to use as medoid
        "order_index",
        "n_emitted_clusters",
        "n_remaining_points",
        # A float value which determines how strictly the clusterer rejects potential clusters.
        # The lower, the stricter. We increase this adaptively to avoid clustering for ever
        "peak_valley_ratio",
        # A deque to store whether the last N candicate clusters were rejected. See `minsuccesses`
        "attempts",
        # An integer storing the number of True values in `attempts` to avoid looping over it
        "successes",
        # A buffer in which a histrogram of distances from the current medoid is stored.
        # Is overwritten on each iteration, we just keep it here to avoid allocations.
        "histogram",
        "histogram_edges",
        # This bool array is False if the contig at the given index has been emitted in a previous iteration.
        # When matrix is on CPU, rows in this array (and the data matrix) is continuously deleted, and we use
        # this array to keep track of which rows to delete each iteration.
        # On GPU, deleting rows is not feasable because that requires us to copy the matrix from and to the GPU,
        # so we instead use this array to mask away any point emitted in an earlier iteration.
        "kept_mask",
        # This dict caches the computation of `sample_medoid`, which is the most compute heavy function
        "medoid_cache",
    ]

    def __repr__(self) -> str:
        return f"ClusterGenerator({len(self.matrix)} points, {self.n_emitted_clusters} clusters)"

    def __str__(self) -> str:
        return f"""ClusterGenerator({len(self.matrix)} points, {self.n_emitted_clusters} clusters)
  CUDA:         {self.cuda}
  maxsteps:     {self.maxsteps}
  minsuccesses: {self.minsuccesses}
  pvr:          {self.peak_valley_ratio}
  successes:    {self.successes}/{len(self.attempts)}
"""

    def _check_params(
        self,
        matrix: _np.ndarray,
        lengths: _np.ndarray,
        maxsteps: int,
        windowsize: int,
        minsuccesses: int,
    ) -> None:
        """Checks matrix, and maxsteps."""

        if matrix.dtype != _np.float32:
            raise (ValueError("Matrix must be of dtype float32"))

        if maxsteps < 1:
            raise ValueError(f"maxsteps must be a positive integer, not {maxsteps}")

        if windowsize < 1:
            raise ValueError(f"windowsize must be at least 1, not {windowsize}")

        if minsuccesses < 1 or minsuccesses > windowsize:
            raise ValueError(
                f"minsuccesses must be between 1 and windowsize, not {minsuccesses}"
            )

        if len(matrix) < 1:
            raise ValueError("Matrix must have at least 1 observation.")

        if len(lengths) != len(matrix):
            raise ValueError("N sequences in lengths and matrix do not match")

    def _init_histogram_kept_mask(self, N: int) -> tuple[_Tensor, _Tensor]:
        "N is number of contigs"

        kept_mask = _torch.ones(N, dtype=_torch.bool)
        if self.cuda:
            kept_mask = kept_mask.cuda()

        histogram = _torch.empty(_ceil(_XMAX / _DELTA_X))
        return histogram, kept_mask

    def __init__(
        self,
        matrix: _np.ndarray,
        lengths: _np.ndarray,
        maxsteps: int = 25,
        windowsize: int = 300,
        minsuccesses: int = 15,
        destroy: bool = False,
        normalized: bool = False,
        cuda: bool = False,
        rng_seed: int = 0,
    ):
        self._check_params(
            matrix,
            lengths,
            maxsteps,
            windowsize,
            minsuccesses,
        )
        if not destroy:
            matrix = matrix.copy()

        torch_matrix = _torch.from_numpy(matrix)
        if not normalized:
            _normalize(torch_matrix, inplace=True)

        # Move to GPU
        torch_lengths = _torch.Tensor(lengths)
        if cuda:
            torch_matrix = torch_matrix.cuda()
            torch_lengths = torch_lengths.cuda()

        self.maxsteps: int = maxsteps
        self.minsuccesses: int = minsuccesses
        self.cuda: bool = cuda
        self.rng: _random.Random = _random.Random(rng_seed)

        self.matrix = torch_matrix
        # This refers to the indices of the original matrix. As we remove points, these
        # indices do not correspond to merely range(len(matrix)) anymore.
        self.indices = _torch.arange(len(matrix))
        self.order = _np.argsort(lengths)[::-1]
        self.order_index = 0
        self.lengths = _torch.Tensor(lengths)
        if cuda:
            self.lengths = self.lengths.cuda()
        self.n_emitted_clusters = 0
        self.n_remaining_points = len(torch_matrix)
        self.peak_valley_ratio = 0.1
        self.attempts: _deque[bool] = _deque(maxlen=windowsize)
        self.successes = 0

        histogram, kept_mask = self._init_histogram_kept_mask(len(self.indices))
        self.histogram = histogram
        self.histogram_edges = _torch.linspace(0.0, _XMAX, round(_XMAX / _DELTA_X) + 1)
        self.kept_mask = kept_mask
        self.medoid_cache: OrderedDict[int, tuple[_Tensor, _Tensor, float]] = (
            OrderedDict()
        )

    # It's an iterator itself
    def __iter__(self):
        return self

    def __next__(self) -> Cluster:
        if self.n_remaining_points == 0:
            raise StopIteration
        assert self.n_remaining_points > 0  # not negative

        cluster, _, points = self.find_cluster()
        self.medoid_cache.clear()
        self.n_emitted_clusters += 1
        self.n_remaining_points -= len(points)

        for point in points:
            self.kept_mask[point] = 0

        # Remove all points that's been clustered away. Is slow it itself, but speeds up
        # distance calculation by having fewer points. Worth it on CPU, not on GPU
        if not self.cuda:
            self.pack()

        return cluster

    def pack(self):
        "Remove all used points from the matrix and indices, and reset kept_mask."
        if self.cuda:
            cpu_kept_mask = self.kept_mask.cpu()
            self.matrix = _vambtools.torch_inplace_maskarray(
                self.matrix.cpu(), cpu_kept_mask
            ).cuda()
            self.indices = self.indices[cpu_kept_mask]

        else:
            _vambtools.torch_inplace_maskarray(self.matrix, self.kept_mask)
            self.indices = self.indices[self.kept_mask]

        self.lengths = self.lengths[self.kept_mask]
        self.kept_mask.resize_(len(self.matrix))
        self.kept_mask[:] = 1
        # The medoid cache is invalidated when the indices shifts
        self.medoid_cache.clear()

    def pack_order(self):
        "Remove all used points from self.order"
        self.order = self.order[self.order > -1]
        assert len(self.order) > 0

    def get_next_seed(self) -> int:
        "Get the next seed index for a new medoid search"
        n_original_contigs = len(self.order)
        i = self.order_index - 1  # we increment by 1 in beginning of loop
        while True:
            # Get the order: That's the original index of the contig.
            i = (i + 1) % n_original_contigs

            # When we loop back to the first index after having passed over all indices before,
            # we potentially have many used up -1 values to skip, so we remove these
            # Since the clustering algorithm may loop over self.order many times, we can potentially
            # save time.
            # When running on GPU, we also take this (rare-ish) chance to pack the on-GPU matrix.
            # This speeds up future distance calculation by making the matrix smaller, but requires
            # moving the matrix from and to GPU, so is slow.
            # Doing it here, which we assume is relatively rarely may be a good compromise
            if i == 0 and self.n_emitted_clusters > 0:
                if self.cuda:
                    self.pack()
                self.pack_order()
                n_original_contigs = len(self.order)

            order = self.order[i]
            # -1 signify an index which has previously been used up
            if order == -1:
                continue

            # Find the new index of this old index
            new_index = cast(int, _torch.searchsorted(self.indices, order).item())
            # It's possible the seed contig `order` is part of a previously emitted cluster,
            # in which case it's not to be found in self.indices. In that case, mark the order
            # as -1, and go to the next one
            if (
                new_index >= len(self.indices)
                or self.indices[new_index].item() != order
                or (self.cuda and not self.kept_mask[new_index].item())
            ):
                self.order[i] = -1
                continue

            # Move to next index for the next time this is called
            self.order_index = i + 1
            return new_index

    def update_successes(self, success: bool):
        """Keeps track of how many clusters have been rejected (False) and accepted (True).
        When sufficiently many False has been seen, the peak_valley_ratio is bumped, which relaxes
        the criteria for rejection.
        This prevents the clusterer from getting stuck in an infinite loop.
        """

        # Keep accurately track of successes if we exceed maxlen
        if len(self.attempts) == self.attempts.maxlen:
            self.successes -= self.attempts.popleft()

        # Add the current success to count
        self.successes += success
        self.attempts.append(success)

        # If less than minsuccesses of the last maxlen attempts were successful,
        # we relax the clustering criteria and reset counting successes.
        if (
            len(self.attempts) == self.attempts.maxlen
            and self.successes < self.minsuccesses
        ):
            self.peak_valley_ratio += 0.1
            self.attempts.clear()
            self.successes = 0

            # After relaxing criteria, start over from the best candidate
            # seed contigs which may have been skipped the first time around
            self.order_index = 0

    def wander_medoid(self, seed) -> tuple[int, _Tensor]:
        """Keeps sampling new points within the cluster until it has sampled
        max_attempts without getting a new set of cluster with lower average
        distance"""

        medoid = seed
        tried: set[int] = {medoid}
        cluster, distances, local_density = self.sample_medoid(seed)
        candidates: list[int] = [i for i in cluster.tolist() if i not in tried]
        candidates = self.rng.sample(candidates, k=min(len(candidates), self.maxsteps))
        i = 0

        while i < len(candidates):
            sampled_medoid = candidates[i]
            tried.add(sampled_medoid)
            sample_cluster, sample_distances, sample_density = self.sample_medoid(
                sampled_medoid
            )

            # If the mean distance of inner points of the sample is lower,
            # we move the medoid and start over
            if sample_density > local_density:
                medoid = sampled_medoid
                distances = sample_distances
                local_density = sample_density
                candidates: list[int] = [
                    i for i in sample_cluster.tolist() if i not in tried
                ]
                candidates = self.rng.sample(
                    candidates, k=min(len(candidates), self.maxsteps)
                )
                i = 0
            else:
                i += 1

        return (medoid, distances)

    def find_threshold(
        self, distances: _Tensor
    ) -> Union[Loner, NoThreshold, tuple[float, float]]:
        # If the point is a loner, immediately return a threshold in where only
        # that point is contained.
        if _torch.count_nonzero(distances < 0.05) == 1:
            return Loner()

        # We need to make a histogram of only the unclustered distances - when run on GPU
        # these have not been removed and we must use the kept_mask
        if self.cuda:
            below_xmax = (distances <= _XMAX) & self.kept_mask
            picked_distances = distances[below_xmax].cpu()
            picked_lengths = self.lengths[below_xmax].cpu()
        else:
            below_xmax = distances <= _XMAX
            picked_distances = distances[below_xmax]
            picked_lengths = self.lengths[below_xmax]

        # TODO: https://github.com/pytorch/pytorch/issues/69519
        # Currently, this function does not run on GPU. This means we must
        # copy over the lengths and distances to CPU each time, which is very slow.
        # If the issue is resolved, there can be large speedups on GPU
        _torch.histogram(
            input=picked_distances,
            bins=len(self.histogram),
            range=(0.0, _XMAX),
            out=((self.histogram, self.histogram_edges)),
            weight=picked_lengths,
        )

        peak_density = 0.0
        peak_over = False
        minimum_x = 0.0
        threshold = None
        delta_x = _XMAX / len(self.histogram)
        pdf_len = len(_NORMALPDF)

        if self.cuda:
            histogram = self.histogram.cpu()
        else:
            histogram = self.histogram

        # This smoothes out the histogram, so we can more reliably detect peaks
        # and valleys.
        densities = _torch.zeros(len(histogram) + pdf_len - 1)
        for i in range(len(densities) - pdf_len + 1):
            densities[i : i + pdf_len] += _NORMALPDF * histogram[i]
        densities = densities[15:-15]

        # Analyze the point densities to find the valley
        x = 0
        density_at_minimum = 0.0
        for density_ in densities:
            density = cast(float, density_.item())
            # Define the first "peak" in point density. That's simply the max until
            # the peak is defined as being over.
            if not peak_over and density > peak_density:
                # Do not accept first peak to be after x = 0.1
                if x > 0.1:
                    return NoThreshold()
                peak_density = density

            # Peak is over when density drops below 60% of peak density
            if not peak_over and density < 0.6 * peak_density:
                peak_over = True
                density_at_minimum = density

            # If another peak is detected, we stop
            if peak_over and density > 1.5 * density_at_minimum:
                break

            # Now find the minimum after the peak
            if peak_over and density < density_at_minimum:
                minimum_x, density_at_minimum = x, density

                # If this minimum is below ratio * peak, it's accepted as threshold
                if density < self.peak_valley_ratio * peak_density:
                    threshold = minimum_x

            x += delta_x

        # If we have not detected a threshold, we can't return one.
        if threshold is None:
            return NoThreshold()
        # Else, we check whether the threshold is too high. If not, we return it.
        else:
            if threshold > 0.2 + self.peak_valley_ratio:
                return NoThreshold()
            else:
                observed_pvr = density_at_minimum / peak_density
                return (threshold, observed_pvr)

    def find_cluster(self) -> tuple[Cluster, int, _Tensor]:
        while True:
            seed = self.get_next_seed()
            medoid, distances = self.wander_medoid(seed)
            threshold = self.find_threshold(distances)
            if isinstance(threshold, Loner):
                cluster = Cluster(
                    int(self.indices[medoid].item()),  # type: ignore
                    seed,
                    _np.array([self.indices[medoid].item()]),
                    self.peak_valley_ratio,
                    None,
                    None,
                    self.successes,
                    len(self.attempts),
                )
                points = _torch.IntTensor([medoid])
                return (cluster, medoid, points)
            elif isinstance(threshold, NoThreshold):
                # When the peak_valley_ratio is too high, we need to return something to not get caught
                # in an infinite loop.
                if self.peak_valley_ratio > 0.55:
                    points = _smaller_indices(
                        distances, self.kept_mask, _DEFAULT_RADIUS, self.cuda
                    )
                    cluster = Cluster(
                        int(self.indices[medoid].item()),  # type: ignore
                        seed,
                        self.indices[points].numpy(),
                        self.peak_valley_ratio,
                        None,
                        _DEFAULT_RADIUS,
                        self.successes,
                        len(self.attempts),
                    )
                    return (cluster, medoid, points)
                else:
                    self.update_successes(False)

            elif isinstance(threshold, tuple):
                (threshold, observed_pvr) = threshold
                points = _smaller_indices(
                    distances, self.kept_mask, threshold, self.cuda
                )
                cluster = Cluster(
                    int(self.indices[medoid].item()),  # type: ignore
                    seed,
                    self.indices[points].numpy(),
                    self.peak_valley_ratio,
                    observed_pvr,
                    threshold,
                    self.successes,
                    len(self.attempts),
                )
                if self.peak_valley_ratio < 0.55:
                    self.update_successes(True)
                return (cluster, medoid, points)

            else:  # No more types
                assert False

    def sample_medoid(
        self,
        medoid: int,
    ) -> tuple[_Tensor, _Tensor, float]:
        """Returns:
        - A vector of indices to points within threshold
        - A vector Xf distances to all points
        - The mean distance from medoid to the other points in the first vector
        """
        existing = self.medoid_cache.get(medoid, None)
        if existing is not None:
            return existing

        distances = _calc_distances(self.matrix, medoid)

        if self.cuda:
            within_threshold = (distances <= _MEDOID_RADIUS) & self.kept_mask
            cluster = _torch.nonzero(within_threshold).flatten().cpu()
        else:
            within_threshold = distances.numpy() <= _MEDOID_RADIUS
            cluster = _torch.from_numpy(within_threshold.nonzero()[0])

        closeness = _MEDOID_RADIUS - distances[within_threshold]
        local_density = (self.lengths[within_threshold] * closeness).sum().item()
        result = (cluster, distances, local_density)

        # Make room for a new element in the cache if full, by dropping the oldest entry
        if len(self.medoid_cache) == MAX_CACHED_RESULTS:
            self.medoid_cache.popitem(last=False)
        self.medoid_cache[medoid] = result

        return result


def _smaller_indices(
    tensor: _Tensor, kept_mask: _Tensor, threshold: float, cuda: bool
) -> _Tensor:
    """Get all indices where the tensor is smaller than the threshold."""

    # If it's on GPU, we remove the already clustered points at this step
    # and move to CPU
    if cuda:
        return _torch.nonzero((tensor <= threshold) & kept_mask).flatten().cpu()
    else:
        return _torch.nonzero(tensor <= threshold).flatten()


def _normalize(matrix: _Tensor, inplace: bool = False) -> _Tensor:
    """Preprocess the matrix to make distance calculations faster.
    The distance functions in this module assumes input has been normalized
    and will not work otherwise.
    """
    if isinstance(matrix, _np.ndarray):
        matrix = _torch.from_numpy(matrix)

    if not inplace:
        matrix = matrix.clone()

    # If any rows are kept all zeros, the distance function will return 0.5 to all points
    # inclusive itself, which can break the code in this module
    zeromask = (matrix == 0).all(dim=1)
    matrix[zeromask] = 1 / matrix.shape[1]
    matrix /= matrix.norm(dim=1).reshape(-1, 1) * (2**0.5)
    return matrix


def _calc_distances(matrix: _Tensor, index: int) -> _Tensor:
    "Return vector of cosine distances from rows of normalized matrix to given row."
    dists = 0.5 - matrix.matmul(matrix[index])
    dists[index] = 0.0  # avoid float rounding errors
    return dists



__doc__ = "Various classes and functions Vamb uses internally."

import gzip as _gzip
import bz2 as _bz2
import lzma as _lzma
import numpy as _np
import re as _re
from vamb._vambtools import _kmercounts, _overwrite_matrix
import collections as _collections
from itertools import zip_longest
from hashlib import md5 as _md5
from collections.abc import Iterable, Iterator
from typing import Optional, IO, Union
from pathlib import Path
from loguru import logger

CLUSTERS_HEADER = "clustername\tcontigname"


def log_and_error(typ: type, msg: str):
    logger.opt(raw=True).info("\n")
    logger.error(msg)
    raise typ(msg)


class BinSplitter:
    """
    The binsplitter can be either
    * Instantiated with an explicit option, in which case `is_default` is False,
      and `splitter` is None if the user explicitly opted out of binsplitting by
      passing in an empty string, else a `str`, the string explicitly asked for
    * Instantiated by default, in which case `is_default` is `True`, and `splitter`
      is `_DEFAULT_SPLITTER`

    The `initialize` function checks the validity of the binsplitter on the set of
    identifiers:
    * If the binsplitter sep is explicitly `None`, do nothing
    * If the binsplitter is default and the separator is not found, warn the user,
      then set the separator to `None` (disabling binsplitting)
    * If the binsplitter is explicitly a string, check this string occurs in all
      identifiers, else error early.
    """

    _DEFAULT_SPLITTER = "C"
    __slots__ = ["is_default", "splitter", "is_initialized"]

    def __init__(self, binsplitter: Optional[str]):
        if binsplitter is None:
            self.is_default = True
            self.splitter = self._DEFAULT_SPLITTER

        else:
            self.is_default = False
            if len(binsplitter) == 0:
                self.splitter = None
            else:
                self.splitter = binsplitter

        self.is_initialized = False

    @classmethod
    def inert_splitter(cls):
        return cls("")

    def initialize(self, identifiers: Iterable[str]):
        if self.is_initialized:
            return None
        self.is_initialized = True
        separator = self.splitter
        if separator is None:
            return None
        message = (
            'Binsplit separator (option `-o`) {imexplicit} passed as "{separator}", '
            'but sequence identifier "{identifier}" does not contain this separator, '
            "or contains it at the very start or end.\n"
            "A binsplit separator X implies that every sequence identifier is formatted as\n"
            "[sample identifier][X][sequence identifier], e.g. a binsplit separator of 'C' "
            "means that 'S1C19' and '7C11' are valid identifiers.\n"
        )

        if not self.is_default:
            for identifier in identifiers:
                (front, _, rest) = identifier.partition(separator)
                if not front or not rest:
                    log_and_error(
                        ValueError,
                        message.format(
                            imexplicit="explicitly",
                            separator=separator,
                            identifier=identifier,
                        ),
                    )
        else:
            for identifier in identifiers:
                (front, _, rest) = identifier.partition(separator)
                if not front or not rest:
                    message += "\nSkipping binsplitting."
                    logger.opt(raw=True).info("\n")
                    logger.warning(
                        message.format(
                            imexplicit="implicitly",
                            separator=separator,
                            identifier=identifier,
                        )
                    )
                    self.splitter = None
                    break

    def split_bin(
        self,
        binname: str,
        identifiers: Iterable[str],
    ) -> Iterable[tuple[str, set[str]]]:
        "Split a single bin by the prefix of the headers"
        if self.splitter is None:
            yield (binname, set(identifiers))
            return None
        else:
            by_sample: dict[str, set[str]] = _collections.defaultdict(set)
            for identifier in identifiers:
                sample, _, rest = identifier.partition(self.splitter)

                if not rest or not sample:
                    raise KeyError(
                        f"Separator '{self.splitter}' not in sequence identifier, or is at the very start or end of identifier: '{identifier}'"
                    )

                by_sample[sample].add(identifier)

            for sample, splitheaders in by_sample.items():
                newbinname = f"{sample}{self.splitter}{binname}"
                yield newbinname, splitheaders

    def binsplit(
        self,
        clusters: Iterable[tuple[str, Iterable[str]]],
    ) -> Iterable[tuple[str, set[str]]]:
        """Splits a set of clusters by the prefix of their names.
        The separator is a string which separated prefix from postfix of contignames. The
        resulting split clusters have the prefix and separator prepended to them.

        clusters can be an iterator, in which case this function returns an iterator, or a dict
        with contignames: set_of_contignames pair, in which case a dict is returned.

        Example:
        >>> clusters = {"bin1": {"s1-c1", "s1-c5", "s2-c1", "s2-c3", "s5-c8"}}
        >>> binsplit(clusters, "-")
        {'s2-bin1': {'s1-c1', 's1-c3'}, 's1-bin1': {'s1-c1', 's1-c5'}, 's5-bin1': {'s1-c8'}}
        """
        for binname, headers in clusters:
            for newbinname, splitheaders in self.split_bin(binname, headers):
                yield newbinname, splitheaders

    def log_string(self) -> str:
        if not self.is_default:
            if self.splitter is None:
                return "Explicitly passed as empty (no binsplitting)"
            else:
                return f'"{self.splitter}"'
        else:
            if self.splitter is None:
                return "Defaulting to 'C', but disabled due to incompatible identifiers"
            else:
                return "Defaulting to 'C'"


class PushArray:
    """Data structure that allows efficient appending and extending a 1D Numpy array.
    Intended to strike a balance between not resizing too often (which is slow), and
    not allocating too much at a time (which is memory inefficient).

    Usage:
    >>> arr = PushArray(numpy.float64)
    >>> arr.append(5.0)
    >>> arr.extend(numpy.linspace(4, 3, 3))
    >>> arr.take() # return underlying Numpy array
    array([5. , 4. , 3.5, 3. ])
    """

    __slots__ = ["data", "capacity", "length"]

    def __init__(self, dtype, start_capacity: int = 1 << 16):
        self.capacity: int = start_capacity
        self.data: _np.ndarray = _np.empty(self.capacity, dtype=dtype)
        self.length = 0

    def __len__(self) -> int:
        return self.length

    def _setcapacity(self, n: int) -> None:
        self.data.resize(n, refcheck=False)
        self.capacity = n

    def _grow(self, mingrowth: int) -> None:
        """Grow capacity by power of two between 1/8 and 1/4 of current capacity, though at
        least mingrowth"""
        growth = max(int(self.capacity * 0.125), mingrowth)
        nextpow2 = 1 << (growth - 1).bit_length()
        self._setcapacity(self.capacity + nextpow2)

    def append(self, value) -> None:
        if self.length == self.capacity:
            self._grow(64)

        self.data[self.length] = value
        self.length += 1

    def extend(self, values) -> None:
        lenv = len(values)
        if self.length + lenv > self.capacity:
            self._grow(lenv)

        self.data[self.length : self.length + lenv] = values
        self.length += lenv

    def take(self) -> _np.ndarray:
        "Return the underlying array"
        self._setcapacity(self.length)
        return self.data

    def clear(self) -> None:
        "Empties the PushArray. Does not clear the underlying memory"
        self.length = 0


def zscore(
    array: _np.ndarray, axis: Optional[int] = None, inplace: bool = False
) -> _np.ndarray:
    """Calculates zscore for an array. A cheap copy of scipy.stats.zscore.

    Inputs:
        array: Numpy array to be normalized
        axis: Axis to operate across [None = entrie array]
        inplace: Do not create new array, change input array [False]

    Output:
        If inplace is True: Input numpy array
        else: New normalized Numpy-array
    """

    if axis is not None and (axis >= array.ndim or axis < 0):
        raise _np.exceptions.AxisError(str(axis))

    if inplace and not _np.issubdtype(array.dtype, _np.floating):
        raise TypeError("Cannot convert a non-float array to zscores")

    mean = array.mean(axis=axis)
    std = array.std(axis=axis)

    if axis is None:
        if std == 0:
            std = 1  # prevent divide by zero

    else:
        std[std == 0.0] = 1  # prevent divide by zero
        shape = tuple(dim if ax != axis else 1 for ax, dim in enumerate(array.shape))
        mean.shape, std.shape = shape, shape

    if inplace:
        array -= mean
        array /= std
        return array
    else:
        return (array - mean) / std


def numpy_inplace_maskarray(array: _np.ndarray, mask: _np.ndarray) -> _np.ndarray:
    """In-place masking of a Numpy array, i.e. if `mask` is a boolean mask of same
    length as `array`, then array[mask] == numpy_inplace_maskarray(array, mask),
    but does not allocate a new array.
    """

    if len(mask) != len(array):
        raise ValueError("Lengths of array and mask must match")
    elif len(array.shape) != 2:
        raise ValueError("Can only take a 2 dimensional-array.")

    uints = _np.frombuffer(mask, dtype=_np.uint8)
    index = _overwrite_matrix(array, uints)
    array.resize((index, array.shape[1]), refcheck=False)
    return array


def torch_inplace_maskarray(array, mask):
    """In-place masking of a Tensor, i.e. if `mask` is a boolean mask of same
    length as `array`, then array[mask] == torch_inplace_maskarray(array, mask),
    but does not allocate a new tensor.
    """

    if len(mask) != len(array):
        raise ValueError("Lengths of array and mask must match")
    elif array.dim() != 2:
        raise ValueError("Can only take a 2 dimensional-array.")

    np_array = array.numpy()
    np_mask = _np.frombuffer(mask.numpy(), dtype=_np.uint8)
    index = _overwrite_matrix(np_array, np_mask)
    array.resize_((index, array.shape[1]))
    return array


def mask_lower_bits(floats: _np.ndarray, bits: int) -> None:
    if bits < 0 or bits > 23:
        raise ValueError("Must mask between 0 and 23 bits")

    mask = ~_np.uint32(2**bits - 1)
    u = floats.view(_np.uint32)
    u &= mask


class Reader:
    """Use this instead of `open` to open files which are either plain text,
    gzipped, bzip2'd or zipped with LZMA.

    Usage:
    >>> with Reader(file, readmode) as file: # by default textmode
    >>>     print(next(file))
    TEST LINE
    """

    def __init__(self, filename: Union[str, Path]):
        self.filename = filename

        with open(self.filename, "rb") as f:
            signature = f.peek(8)[:8]

        # Gzipped files begin with the two bytes 0x1F8B
        if tuple(signature[:2]) == (0x1F, 0x8B):
            self.filehandle = _gzip.open(self.filename, "rb")

        # bzip2 files begin with the signature BZ
        elif signature[:2] == b"BZ":
            self.filehandle = _bz2.open(self.filename, "rb")

        # .XZ files begins with 0xFD377A585A0000
        elif tuple(signature[:7]) == (0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00, 0x00):
            self.filehandle = _lzma.open(self.filename, "rb")

        # Else we assume it's a text file.
        else:
            self.filehandle = open(self.filename, "rb")

    def close(self):
        self.filehandle.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, _value, _traceback):
        self.close()

    def __iter__(self):
        return self.filehandle


class FastaEntry:
    """One single FASTA entry. Instantiate with byte header and bytearray
    sequence."""

    # IUPAC ambiguous DNA letters + u
    allowed = b"acgtuswkmyrbdhvn"
    allowed += allowed.upper()
    # Allow only the same identifier chars that can be in the BAM file, otherwise
    # users will be frustrated with FASTA and BAM headers do not match.
    # BAM only includes identifier, e.g. stuff before whitespace. So we accept anything
    # after whitespace, but do not use it as header.
    # This brutal regex is derived from the SAM specs, valid identifiers,
    # but disallow leading # sign, and allow trailing whitespace + ignored description
    regex = _re.compile(
        b"([0-9A-Za-z!$%&+./:;?@^_|~-][0-9A-Za-z!#$%&*+./:;=?@^_|~-]*)([^\\S\r\n][^\r\n]*)?$"
    )
    __slots__ = ["identifier", "description", "sequence"]

    def _verify_header(self, header: bytes) -> tuple[str, str]:
        m = self.regex.match(header)
        if m is None:
            raise ValueError(
                f'Invalid header in FASTA: "{header.decode()}". '
                '\nMust conform to identifier regex pattern of SAM specification: "'
                '>([0-9A-Za-z!$%&+./:;?@^_|~-][0-9A-Za-z!#$%&*+./:;=?@^_|~-]*)([^\\S\\r\\n][^\\r\\n]*)?$".\n'
                "If the header does not fit this pattern, the header cannot be represented in BAM files, "
                "which means Vamb cannot compare sequences in BAM and FASTA files."
            )
        identifier, description = m.groups()
        description = "" if description is None else description.decode()
        return (identifier.decode(), description)

    def __init__(self, header: bytes, sequence: bytearray):
        identifier, description = self._verify_header(header)
        self.identifier: str = identifier
        self.description: str = description
        masked = sequence.translate(None, b" \t\n\r")
        stripped = masked.translate(None, self.allowed)
        if len(stripped) > 0:
            codeunit = stripped[0]
            bad_character = chr(codeunit)
            raise ValueError(
                f"Non-IUPAC DNA/RNA byte in sequence {identifier}: '{bad_character}', byte value {codeunit}"
            )

        self.sequence: bytearray = masked

    @property
    def header(self) -> str:
        return self.identifier + self.description

    def rename(self, header: bytes) -> None:
        identifier, description = self._verify_header(header)
        self.identifier = identifier
        self.description = description

    def __len__(self) -> int:
        return len(self.sequence)

    def format(self, width: int = 60) -> str:
        sixtymers = range(0, len(self.sequence), width)
        spacedseq = "\n".join(
            [self.sequence[i : i + width].decode() for i in sixtymers]
        )
        return f">{self.header}\n{spacedseq}"

    def kmercounts(self, k: int) -> _np.ndarray:
        if k < 1 or k > 10:
            raise ValueError("k must be between 1 and 10 inclusive")

        counts = _np.zeros(1 << (2 * k), dtype=_np.int32)
        _kmercounts(self.sequence, k, counts)
        return counts


def byte_iterfasta(
    filehandle: Iterable[bytes], comment: bytes = b"#"
) -> Iterator[FastaEntry]:
    """Yields FastaEntries from a binary opened fasta file.

    Usage:
    >>> with Reader('/dir/fasta.fna') as filehandle:
    ...     entries = byte_iterfasta(filehandle) # a generator

    Inputs:
        filehandle: Any iterator of binary lines of a FASTA file
        comment: Ignore lines beginning with any whitespace + comment

    Output: Generator of FastaEntry-objects from file
    """

    # Make it work for persistent iterators, e.g. lists
    line_iterator = iter(filehandle)
    # Skip to first header
    try:
        for probeline in line_iterator:
            stripped = probeline.lstrip()
            if stripped.startswith(comment):
                pass

            elif probeline[0:1] == b">":
                break

            else:
                raise ValueError("First non-comment line is not a Fasta header")

        else:  # no break
            return None
            # raise ValueError('Empty or outcommented file')

    except TypeError:
        errormsg = (
            "First line does not contain bytes. Are you reading file in binary mode?"
        )
        raise TypeError(errormsg) from None

    # 13 is the byte value of \r, meaning we remove either \r\n or \n
    header = probeline[1 : -(1 + (probeline[-2] == 13))]
    buffer: list[bytes] = list()

    # Iterate over lines
    for line in line_iterator:
        if line.startswith(comment):
            pass

        elif line.startswith(b">"):
            yield FastaEntry(header, bytearray().join(buffer))
            buffer.clear()
            header = line[1 : -(1 + (line[-2] == 13))]

        else:
            buffer.append(line)

    yield FastaEntry(header, bytearray().join(buffer))


class RefHasher:
    __slots__ = ["hasher"]

    def __init__(self):
        self.hasher = _md5()

    def add_refname(self, ref: str) -> None:
        self.hasher.update(ref.encode().rstrip())

    def add_refnames(self, refs: Iterable[str]):
        for ref in refs:
            self.add_refname(ref)
        return self

    @classmethod
    def hash_refnames(cls, refs: Iterable[str]) -> bytes:
        return cls().add_refnames(refs).digest()

    def digest(self) -> bytes:
        return self.hasher.digest()

    @staticmethod
    def verify_refhash(
        refhash: bytes,
        target_refhash: bytes,
        observed_name: Optional[str],
        target_name: Optional[str],
        identifiers: Optional[tuple[Iterable[str], Iterable[str]]],
    ) -> None:
        if refhash == target_refhash:
            return None

        obs_name = "observed" if observed_name is None else observed_name
        tgt_name = "target" if target_name is None else target_name

        message = (
            f"Mismatch between sequence identifiers (names) in {obs_name} and {tgt_name}.\n"
            f"Observed {obs_name} identifier hash: {refhash.hex()}.\n"
            f"Expected {tgt_name} identifier hash: {target_refhash.hex()}\n"
            f"Make sure all identifiers in {obs_name} and {tgt_name} are identical "
            "and in the same order. "
            "Note that the identifier is the header before any whitespace."
        )

        if identifiers is not None:
            (observed_ids, target_ids) = identifiers
            for i, (observed_id, target_id) in enumerate(
                zip_longest(observed_ids, target_ids)
            ):
                if observed_id is None:
                    message += (
                        f"\nIdentifier mismatch: {obs_name} has only "
                        f"{i} identifier(s), which is fewer than {tgt_name}"
                    )
                    log_and_error(ValueError, message)
                elif target_id is None:
                    message += (
                        f"\nIdentifier mismatch: {tgt_name} has only "
                        f"{i} identifier(s), which is fewer than {obs_name}"
                    )
                    log_and_error(ValueError, message)
                elif observed_id != target_id:
                    message += (
                        f"\nIdentifier mismatch: Identifier number {i + 1} does not match "
                        f"between {obs_name} and {tgt_name}:"
                        f'{obs_name}: "{observed_id}"'
                        f'{tgt_name}: "{target_id}"'
                    )
                    log_and_error(ValueError, message)
            assert False
        else:
            log_and_error(ValueError, message)


def write_clusters(
    io: IO[str],
    clusters: Iterable[tuple[str, set[str]]],
) -> tuple[int, int]:
    n_clusters = 0
    n_contigs = 0
    print(CLUSTERS_HEADER, file=io)
    for cluster_name, contig_names in clusters:
        n_clusters += 1
        n_contigs += len(contig_names)
        for contig_name in contig_names:
            print(cluster_name, contig_name, sep="\t", file=io)

    return (n_clusters, n_contigs)


def read_clusters(filehandle: Iterable[str], min_size: int = 1) -> dict[str, set[str]]:
    """Read clusters from a file as created by function `writeclusters`.

    Inputs:
        filehandle: An open filehandle that can be read from
        min_size: Minimum number of contigs in cluster to be kept

    Output: A {clustername: set(contigs)} dict"""

    contigsof: _collections.defaultdict[str, set[str]] = _collections.defaultdict(set)
    lines = iter(filehandle)
    header = next(lines)
    if header.rstrip(" \n") != CLUSTERS_HEADER:
        raise ValueError(
            f'Expected cluster TSV file to start with header: "{CLUSTERS_HEADER}"'
        )

    for line in lines:
        stripped = line.strip()

        if not stripped or stripped[0] == "#":
            continue

        clustername, contigname = stripped.split("\t")
        contigsof[clustername].add(contigname)

    contigsof_dict = {cl: co for cl, co in contigsof.items() if len(co) >= min_size}

    return contigsof_dict


def check_is_creatable_file_path(path: Path) -> None:
    if path.exists():
        raise FileExistsError(path)
    if not path.parent.is_dir():
        raise NotADirectoryError(path.parent)


def create_dir_if_not_existing(path: Path) -> None:
    if path.is_dir():
        return None
    if path.is_file():
        raise FileExistsError(path)
    if not path.parent.is_dir():
        raise NotADirectoryError(path.parent)
    path.mkdir(exist_ok=True)


def write_bins(
    directory: Path,
    bins: dict[str, set[str]],
    fastaio: Iterable[bytes],
    maxbins: Optional[int] = 1000,
):
    """Writes bins as FASTA files in a directory, one file per bin.

    Inputs:
        directory: Directory to create or put files in
        bins: dict[str: set[str]] (can be loaded from clusters.tsv using vamb.cluster.read_clusters)
        fastaio: bytes iterator containing FASTA file with all sequences
        maxbins: None or else raise an error if trying to make more bins than this [1000]
    Output: None
    """

    # Safety measure so someone doesn't accidentally make 50000 tiny bins
    # If you do this on a compute cluster it can grind the entire cluster to
    # a halt and piss people off like you wouldn't believe.
    if maxbins is not None and len(bins) > maxbins:
        raise ValueError(f"{len(bins)} bins exceed maxbins of {maxbins}")

    create_dir_if_not_existing(directory)

    keep: set[str] = set()
    for i in bins.values():
        keep.update(i)

    bytes_by_id: dict[str, bytes] = dict()
    for entry in byte_iterfasta(fastaio):
        if entry.identifier in keep:
            bytes_by_id[entry.identifier] = _gzip.compress(
                entry.format().encode(), compresslevel=1
            )

    # Now actually print all the contigs to files
    for binname, contigs in bins.items():
        for contig in contigs:
            byts = bytes_by_id.get(contig)
            if byts is None:
                raise IndexError(
                    f'Contig "{contig}" in bin missing from input FASTA file'
                )

        # Print bin to file
        with open(directory.joinpath(binname + ".fna"), "wb") as file:
            for contig in contigs:
                file.write(_gzip.decompress(bytes_by_id[contig]))
                file.write(b"\n")


def validate_input_array(array: _np.ndarray) -> _np.ndarray:
    "Returns array similar to input array but C-contiguous and with own data."
    if not array.flags["C_CONTIGUOUS"]:
        array = _np.ascontiguousarray(array)
    if not array.flags["OWNDATA"]:
        array = array.copy()

    assert array.flags["C_CONTIGUOUS"] and array.flags["OWNDATA"]
    return array


def read_npz(file) -> _np.ndarray:
    """Loads array in .npz-format

    Input: Open file or path to file with npz-formatted array

    Output: A Numpy array
    """

    npz = _np.load(file)
    array = validate_input_array(npz["arr_0"])
    npz.close()

    return array


def write_npz(file, array: _np.ndarray):
    """Writes a Numpy array to an open file or path in .npz format

    Inputs:
        file: Open file or path to file
        array: Numpy array

    Output: None
    """
    _np.savez_compressed(file, array)


def concatenate_fasta_ios(
    outfile: IO[str],
    readers: Iterable[Iterable[bytes]],
    minlength: int = 2000,
    rename: bool = True,
):
    """Creates a new FASTA file from input paths, and optionally rename contig headers
    to the pattern "S{sample number}C{contig identifier}".

    Inputs:
        outpath: Open filehandle for output file
        readers: Iterable of iterable of bytes to read from, representing FASTA sequences
        minlength: Minimum contig length to keep [2000]
        rename: Rename headers

    Output: None
    """

    identifiers: set[str] = set()
    for reader_no, reader in enumerate(readers):
        # If we rename, seq identifiers only have to be unique for each sample
        if rename:
            identifiers.clear()

        for entry in byte_iterfasta(reader):
            if len(entry) < minlength:
                continue

            if rename:
                entry.rename(f"S{reader_no + 1}C{entry.identifier}".encode())

            if entry.identifier in identifiers:
                raise ValueError(
                    "Multiple sequences would be given "
                    f'identifier "{entry.identifier}".'
                )
            identifiers.add(entry.identifier)
            print(entry.format(), file=outfile)


def concatenate_fasta(
    outfile: IO[str],
    inpaths: Iterable[Path],
    minlength: int = 2000,
    rename: bool = True,
):
    concatenate_fasta_ios(
        outfile, open_file_iterator(inpaths), minlength=minlength, rename=rename
    )


def open_file_iterator(paths: Iterable[Path]) -> Iterable[Reader]:
    for path in paths:
        with Reader(path) as io:
            yield io



import string
import os
import pathlib

import vamb

PARENTDIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DATADIR = os.path.join(PARENTDIR, "test", "data")
BAM_FILES = sorted(
    [
        pathlib.Path(DATADIR).joinpath("bam").joinpath(i)
        for i in os.listdir(os.path.join(DATADIR, "bam"))
    ]
)
AEMB_DIR = os.path.join(DATADIR, "aemb")
AEMB_FILES = sorted([pathlib.Path(AEMB_DIR).joinpath(i) for i in os.listdir(AEMB_DIR)])

BAM_NAMES = [
    "S27C175628",
    "S27C95602",
    "S27C25358",
    "S26C115410",
    "S4C529736",
    "S27C181335",
    "S4C222286",
    "S27C38468",
    "S11C13125",
    "S4C480978",
    "S27C255582",
    "S27C170328",
    "S7C221395",
    "S26C281881",
    "S12C228927",
    "S26C86604",
    "S27C93037",
    "S9C124493",
    "S27C236159",
    "S27C214882",
    "S7C273086",
    "S8C93079",
    "S12C85159",
    "S10C72456",
    "S27C19079",
]

BAM_SEQ_LENS = [
    2271,
    3235,
    3816,
    2625,
    2716,
    4035,
    3001,
    2583,
    5962,
    3774,
    2150,
    2161,
    2218,
    2047,
    5772,
    2633,
    3400,
    3502,
    2103,
    4308,
    3061,
    2464,
    4099,
    2640,
    2449,
]


def make_randseq(rng, frm: int, to: int) -> vamb.vambtools.FastaEntry:
    name = rng.choice(string.ascii_uppercase) + "".join(
        rng.choices(string.ascii_lowercase, k=11)
    )
    seq = "".join(
        rng.choices(
            "acgtACGTnNywsdbK",
            weights=[0.12] * 8 + [0.005] * 8,
            k=rng.randrange(frm, to),
        )
    )
    return vamb.vambtools.FastaEntry(name.encode(), bytearray(seq.encode()))



#cython: language_level=3, boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False

cpdef int _overwrite_matrix(float[:,::1] matrix, unsigned char[::1] mask):
    """Given a float32 matrix and Uint8 mask, does the same as setting the first
    rows of matrix to matrix[mask], but in-place.
    This is only important to save on memory.
    """

    cdef int i = 0
    cdef int j = 0
    cdef int matrixindex = 0
    cdef int length = matrix.shape[1]
    cdef int masklength = len(mask)

    # First skip to the first zero in the mask, since the matrix at smaller
    # indices than this should remain untouched.
    for i in range(masklength):
        if mask[i] == 0:
            break

    # If the mask is all true, don't touch array.
    if i == masklength:
        return masklength

    matrixindex = i

    for i in range(matrixindex, masklength):
        if mask[i] == 1:
            for j in range(length):
                matrix[matrixindex, j] = matrix[i, j]
            matrixindex += 1

    return matrixindex

cpdef void _kmercounts(unsigned char[::1] bytesarray, int k, int[::1] counts):
    """Count tetranucleotides of contig and put them in counts vector.

    The bytearray is expected to be np.uint8 of bytevalues of the contig.
    Only values 64, 67, 71, 84 are accepted, all others are skipped.
    The counts is expected to be an array of 4^k 32-bit integers with value 0.
    """

    cdef unsigned int kmer = 0
    cdef int character, charvalue, i
    cdef int countdown = k-1
    cdef int contiglength = len(bytesarray)
    cdef int mask = (1 << (2 * k)) - 1
    cdef unsigned char* lut = [
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 0, 4, 1, 4, 4, 4, 2, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 0, 4, 1, 4, 4, 4, 2, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
        4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    ]

    for i in range(contiglength):
        character = bytesarray[i]
        charvalue = lut[character]

        if charvalue == 4:
            countdown = k

        kmer = ((kmer << 2) | charvalue) & mask

        if countdown == 0:
            counts[kmer] += 1
        else:
            countdown -= 1



import os
import numpy as np
import argparse
import vamb
import shutil
from ordered_set import OrderedSet
import json


def main(
    path_run,
    cluster_contigs,
    cluster_scores,
    cluster_not_r_contigs,
    ripped_bins_scores_ar,
    drep_folder,
    bin_path,
    path_bins_ripped,
    bin_separator,
    min_comp,
    max_cont,
):
    # {cluster:sample} dictionary for the manually drep clusters
    cluster_sample = get_cluster_sample(cluster_contigs, bin_separator)
    for sample in set(cluster_sample.values()):
        os.makedirs(os.path.join(path_run, drep_folder, sample), exist_ok=True)

    # Given all the clusters that have no connection (no intersection) with any other cluster
    # mv them to the final bins folder that contains the final set of NC bins if they are NC
    nc_clusters_unchanged = mv_nc_not_r_nc_bins(
        cluster_not_r_contigs,
        cluster_sample,
        cluster_scores,
        drep_folder,
        bin_path,
        min_comp,
        max_cont,
    )

    # Given the checkM2 scores for all ripped bins, create a dictionary
    # {ripped_bin:[comp,cont]} and also the ripped_bins pairs that share the interseciton
    cluster_r_scores, cluster_r_pairs = get_cluster_r_scores(ripped_bins_scores_ar)

    # Now, given the ripped_bins scores and the pairs, for each pair select one ripped and one original
    # per pair, and move them to the final NC folder
    nc_clusters_unchanged_2, nc_clusters_ripped = choose_best_ripped_bin_and_mv_if_nc(
        cluster_r_scores,
        cluster_scores,
        cluster_r_pairs,
        cluster_sample,
        bin_path,
        drep_folder,
        path_run,
        path_bins_ripped,
        min_comp,
        max_cont,
    )

    # for the bins that have been ripped because of meaningless edges and after the ripping present
    # no intersection with any other bin, also move them to the final set of NC bins if
    # they are NC
    nc_clusters_ripped_single = mv_single_ripped_nc_bins(
        ripped_bins_scores_ar,
        drep_folder,
        cluster_r_pairs,
        bin_path,
        cluster_scores,
        path_run,
        cluster_sample,
        min_comp,
        max_cont,
    )

    nc_clusters_unchanged = nc_clusters_unchanged.union(nc_clusters_unchanged_2)
    nc_clusters_ripped = nc_clusters_ripped.union(nc_clusters_ripped_single)
    quality_file = os.path.join(path_run, drep_folder, "quality_report.tsv")
    write_nc_bin_scores(
        cluster_scores,
        cluster_r_scores,
        nc_clusters_unchanged,
        nc_clusters_ripped,
        quality_file,
    )


def get_cluster_sample(cluster_contigs, bin_separator):
    """{cluster:sample} for all clusters"""
    cluster_sample = {}

    for cluster, contigs in cluster_contigs.items():
        contig_i = next(iter(contigs))
        sample = contig_i.split(bin_separator)[0]

        cluster_sample[cluster] = sample
    return cluster_sample


def mv_nc_not_r_nc_bins(
    cluster_not_r_contigs,
    cluster_sample,
    cluster_scores,
    drep_folder,
    bin_path,
    min_comp,
    max_cont,
):
    """mv not ripped NC bins from bins folder to drep folder"""
    nc_clusters_unchanged = set()
    for cluster in cluster_not_r_contigs.keys():
        comp, cont = cluster_scores[cluster]
        if comp >= min_comp and cont <= max_cont:
            # src_bin=os.path.join(path_bins,cluster_sample[cluster],cluster+'.fna')
            src_bin = bin_path[cluster + ".fna"]
            trg_bin = os.path.join(
                path_run, drep_folder, cluster_sample[cluster], cluster + ".fna"
            )
            print(
                "Bin %s has no intersection with any other bin so it is directly moved from %s to %s"
                % (cluster, src_bin, trg_bin)
            )
            shutil.move(src_bin, trg_bin)

            nc_clusters_unchanged.add(cluster)
    return nc_clusters_unchanged


# For ripped bins


def mv_single_ripped_nc_bins(
    ripped_bins_scores_ar,
    drep_folder,
    cluster_r_pairs,
    # path_bins_ripped,
    bin_path,
    cluster_score,
    path_run,
    cluster_sample,
    min_comp,
    max_cont,
):
    """Mv nc bins that were ripped only because they had meaningless edges"""
    nc_clusters_ripped_single = set()
    clusters_alredy_moved = set(
        [element for pair in cluster_r_pairs for element in pair]
    )
    for row in ripped_bins_scores_ar:
        cluster, comp, cont = row[:3]
        if "--" in cluster or cluster in clusters_alredy_moved:
            continue

        comp, cont = float(comp), float(cont)
        comp_, cont_ = cluster_score[cluster]

        assert comp == comp_
        assert cont == cont_

        if comp >= min_comp and cont <= max_cont:
            src_bin = bin_path[cluster + ".fna"]

            if os.path.isfile(src_bin):
                trg_bin = os.path.join(
                    path_run, drep_folder, cluster_sample[cluster], cluster + ".fna"
                )
                print(
                    "Bin %s was ripped because of meaningless edges or pairing and afterwards no intersection was shared with any other bin so it is moved from %s to %s"
                    % (cluster, src_bin, trg_bin)
                )
                shutil.move(src_bin, trg_bin)

                nc_clusters_ripped_single.add(cluster)
    return nc_clusters_ripped_single


def get_cluster_r_scores(ripped_bins_scores_ar):
    """{cluster:[comp,cont]} for ripped clusters"""

    cluster_r_pairs = []
    cluster_ripped_scores_dict = dict()

    for row in ripped_bins_scores_ar:
        bin_A_bin_B_name, comp, cont = row[:3]
        if "--" in bin_A_bin_B_name:
            bin_A_name, bin_B_name = bin_A_bin_B_name.split("--")
            cluster_A_name, cluster_B_name = bin_A_name, bin_B_name

            cluster_ripped_scores_dict[cluster_A_name] = [float(comp), float(cont)]

            cluster_r_pairs.append(OrderedSet([cluster_A_name, cluster_B_name]))
        else:
            cluster_A_name = bin_A_bin_B_name.replace(".fna", "")
            cluster_ripped_scores_dict[cluster_A_name] = [float(comp), float(cont)]

    return cluster_ripped_scores_dict, cluster_r_pairs


def choose_best_ripped_bin_and_mv_if_nc(
    cluster_r_scores,
    cluster_scores,
    cluster_r_pairs,
    cluster_sample,
    bin_path,
    drep_folder,
    path_run,
    path_bins_ripped,
    min_comp,
    max_cont,
):
    nc_clusters_unchanged, nc_clusters_ripped = set(), set()
    clusters_already_processed = set()
    """So given 2 ripped bins that miss the interseciton, we have to choose who keeps the contigs"""
    for cluster_A_r, cluster_B_r in cluster_r_pairs:
        if (
            cluster_A_r in clusters_already_processed
            or cluster_B_r in clusters_already_processed
        ):
            continue
        print("Bin %s and bin %s share some contigs " % (cluster_A_r, cluster_B_r))

        cluster_A_complet, cluster_A_cont = cluster_scores[cluster_A_r]
        cluster_A_r_complet, cluster_A_r_cont = cluster_r_scores[cluster_A_r]
        cluster_A_score = cluster_A_complet - 5 * cluster_A_cont
        cluster_A_r_score = cluster_A_r_complet - 5 * cluster_A_r_cont

        cluster_B_complet, cluster_B_cont = cluster_scores[cluster_B_r]
        cluster_B_r_complet, cluster_B_r_cont = cluster_r_scores[cluster_B_r]
        cluster_B_score = cluster_B_complet - 5 * cluster_B_cont
        cluster_B_r_score = cluster_B_r_complet - 5 * cluster_B_r_cont

        cluster_B_dif = cluster_B_score - cluster_B_r_score
        cluster_A_dif = cluster_A_score - cluster_A_r_score

        keeper = np.argmax([cluster_A_dif, cluster_B_dif])

        if keeper == 0:
            print(
                "%s keeps the intersecting contigs whereas %s does not "
                % (cluster_A_r, cluster_B_r)
            )

            # bin_A keeps the contigs
            if cluster_A_complet >= min_comp and cluster_A_cont <= max_cont:
                bin_A_name = cluster_A_r + ".fna"
                src_bin = bin_path[bin_A_name]
                trg_bin = os.path.join(
                    path_run, drep_folder, cluster_sample[cluster_A_r], bin_A_name
                )
                shutil.move(src_bin, trg_bin)
                nc_clusters_unchanged.add(cluster_A_r)
                print("%s keeps the contigs so src_path is %s " % (bin_A_name, src_bin))
            # and bin B not
            if cluster_B_r_complet >= min_comp and cluster_B_r_cont <= max_cont:
                bin_B_ripped_name = cluster_B_r + "--" + cluster_A_r + ".fna"
                bin_B_name = cluster_B_r + ".fna"
                src_bin = os.path.join(path_bins_ripped, bin_B_ripped_name)
                trg_bin = os.path.join(
                    path_run, drep_folder, cluster_sample[cluster_B_r], bin_B_name
                )
                shutil.move(src_bin, trg_bin)
                nc_clusters_ripped.add(cluster_B_r)
                print(
                    "%s looses the contigs so src_path is %s " % (bin_B_name, src_bin)
                )

        if keeper == 1:
            print(
                "%s keeps the intersecting contigs whereas %s does not "
                % (cluster_B_r, cluster_A_r)
            )

            # bin_B keeps the contigs
            if cluster_B_complet >= min_comp and cluster_B_cont <= max_cont:
                bin_B_name = cluster_B_r + ".fna"
                src_bin = bin_path[bin_B_name]
                trg_bin = os.path.join(
                    path_run, drep_folder, cluster_sample[cluster_B_r], bin_B_name
                )
                shutil.move(src_bin, trg_bin)
                nc_clusters_unchanged.add(cluster_B_r)
                print("%s keeps the contigs so src_path is %s " % (bin_B_name, src_bin))
            # and bin A not
            if cluster_A_r_complet >= min_comp and cluster_A_r_cont <= max_cont:
                bin_A_ripped_name = cluster_A_r + "--" + cluster_B_r + ".fna"
                bin_A_name = cluster_A_r + ".fna"
                src_bin = os.path.join(path_bins_ripped, bin_A_ripped_name)
                trg_bin = os.path.join(
                    path_run, drep_folder, cluster_sample[cluster_A_r], bin_A_name
                )
                shutil.move(src_bin, trg_bin)
                nc_clusters_ripped.add(cluster_A_r)
                print(
                    "%s looses the contigs so src_path is %s " % (bin_A_name, src_bin)
                )

        clusters_already_processed.add(cluster_A_r)
        clusters_already_processed.add(cluster_B_r)

    return nc_clusters_unchanged, nc_clusters_ripped


def write_nc_bin_scores(
    cluster_scores,
    cluster_r_scores,
    nc_clusters_unchanged,
    nc_clusters_ripped,
    quality_file,
):
    """Given that we know which bins were ripped and which not, write a final
    quality_score.tsv file that collect comp and cont for all NC final bins"""
    with open(quality_file, "a") as file_:
        print("Name", "completeness", "contamination", sep="\t", file=file_)
        file_.flush()
        for nc_cluster in nc_clusters_unchanged:
            comp, cont = cluster_scores[nc_cluster]
            print(nc_cluster, comp, cont, sep="\t", file=file_)
            file_.flush()
        for nc_cluster_r in nc_clusters_ripped:
            comp, cont = cluster_r_scores[nc_cluster_r]
            print(nc_cluster_r, comp, cont, sep="\t", file=file_)
            file_.flush()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-r", type=str, help="path run generated by AVAMB")
    parser.add_argument("--c", type=str, help="path clusters file")
    parser.add_argument("--cnr", type=str, help="path dereplicated clusters non ripped")
    parser.add_argument("--sbr", type=str, help="path scores ripped bins ")
    parser.add_argument("--cs_d", type=str, help="cluster_scores dictionary path ")
    parser.add_argument("--bp_d", type=str, help="bin_path dictionary path")
    parser.add_argument("--br", type=str, help="path bins ripped ")
    parser.add_argument(
        "-d", type=str, help="folder name that will contain all nc bins"
    )
    parser.add_argument("--bin_separator", type=str, help="path ripped bins")
    parser.add_argument("--comp", type=float, default=0.9, help="Min completeness ")
    parser.add_argument("--cont", type=float, default=0.05, help="Max contamination ")

    opt = parser.parse_args()

    path_run = opt.r

    path_clusters = opt.c
    with open(path_clusters) as file:
        cluster_contigs = vamb.vambtools.read_clusters(file)

    with open(opt.cs_d) as f:
        cluster_scores = json.load(f)

    path_clusters_not_r = opt.cnr

    with open(path_clusters_not_r) as file:
        cluster_not_r_contigs = vamb.vambtools.read_clusters(file)

    path_bins_ripped_checkm_quality_report = opt.sbr

    ripped_bins_scores_ar = np.loadtxt(
        path_bins_ripped_checkm_quality_report, dtype=str, skiprows=1, ndmin=2
    )

    drep_folder = opt.d

    with open(opt.bp_d) as f:
        bin_path = json.load(f)

    path_bins_ripped = opt.br

    bin_separator = opt.bin_separator

    main(
        path_run,
        cluster_contigs,
        cluster_scores,
        cluster_not_r_contigs,
        ripped_bins_scores_ar,
        drep_folder,
        bin_path,
        path_bins_ripped,
        bin_separator,
        opt.comp * 100,
        opt.cont * 100,
    )
